/*
                           ***   WARNING   ***

    These functions assume that functions like toupper() are implemted as
    FUNCTIONS, not MACROS!

    Called with embedded increments, when compiled with MACROS the results
    will be wrong!
*/

#define TRUE 1
#define FALSE 0
#define SYM_TABLE_SIZE 1024
#define LABEL_LEN 80
#define LINE_LEN 512

#define ERR -1  /* Err return for file open */
#define MODE_READ 0  /* for call to open() */
#define READ_WRITE 2  /* for call to open() */

#define ERR_NONE          0
#define ERR_EOF           1
             /* this is a reason to stop, but not to print an error message */
#define ERR_NOMATCH_LABEL 2
#define ERR_NOMATCH_INSTR 3
#define ERR_BAD_READ_COM  4
#define ERR_BAD_READ_CG   5
#define ERR_WRITE_ERROR   6
#define ERR_CG_FILE_EOF   7

struct symbol_table
   {
   unsigned int ofs;   /* offset of the label */
   unsigned int seg;   /* segment of the label */
   unsigned char sym[32];  /* and a 31-byte (plus terminator) symbol */
   };

struct symbol_table symbols[SYM_TABLE_SIZE];

unsigned int num_syms;  /* how many symbols are used */
unsigned int parse_seg;
unsigned int is_exefile;  /* is .EXE or .COM */
unsigned int reg_cs;
unsigned int reg_ss;
unsigned int reg_ip;  /* load offset */
unsigned int tab_spacing = 48;  /* how far to tabout comments */

unsigned char globalstr[150];  /* global temporary string */
unsigned char sym_tab_full_msg[] = "Symbol table full.\n";
unsigned char cg_line[LINE_LEN];  /* a big buffer */
unsigned char com_line[LINE_LEN];  /* another big buffer */
unsigned char templine[LINE_LEN];
unsigned char templine2[LINE_LEN];  /* some temp stuff */
unsigned char newline[LINE_LEN];  /* a buffer for line replacement */
unsigned char nl_str[] = "\n";
unsigned char tab_str[] = "\t";
unsigned char copyright[] = "Copyright (c) 1988 by William Couture.";

/*
    NOTE:  On the compiler I'm using (DeSmet, v2.51), stream files
           (fopen(), fread(), fclose(), etc.) return an int.  These
           declarations may (almost certainly will!) be changed for
           your compiler (probably int will change to FILE *).

           Also, just so you know:  The DeSmet puts() function does NOT
           append a newline ('\n') to the output string (a feature that
           I like)!  For another compiler like Microsoft, the equivalent
           function would be cputs().
*/
int commented_file, codegen_file, output_file, symbol_file;
int executable_file;  /* used with open(), read(), close(), etc.
                         will remain an int on your compiler! */

void itoa(number,str)    /* write what should be a library routine */
   int number;
   char *str;
   {
   char buffer[7], *ptr;

   buffer[6] = '\0';
   ptr = &buffer[5];
   do
      {
      *(ptr--) = number  % 10 + '0';  /* count backward rather than use */
      } while ((number /= 10) > 0);   /* a string reverse function */
   ptr++;
   strcpy(str, ptr);
   }

char *strupr(str)
   char *str;
   {
   int i;

   i = -1;
   while (str[++i])
      str[i] = toupper(str[i]);
   return(str);
   }

int compare_sym_addr(first,second)  /* for QSORT routine */
   struct symbol_table *first, *second;
   {
   unsigned int retval;

   if (first->seg != second->seg)
      return(first->seg > second->seg ? 1 : -1);
   if (first->ofs != second->ofs)
      return(first->ofs > second->ofs ? 1 : -1);
   return(0);
   }

int parse_symbol()  /* string to parse passed in globalstr */
   {
   unsigned int hold1, hold2, ptr, i, temp;
   unsigned char ch,byte;

   ptr = 0;
   parse_seg = 0;
   if ((parse_seg_and_ofs(&hold1,&hold2,&ptr,globalstr)) && (globalstr[ptr-1] == ' '))
      {
      for (i = 0; i < 31; i++)
         {
         ch = globalstr[ptr+i];
         if ((!ch) || (ch == '\015') || (ch == '\012'))
            break;
         }
      if (i)
         {
         globalstr[ptr+i] = '\0';  /* terminate at CR or 31 chars */
         symbols[num_syms].ofs = hold2;
         symbols[num_syms].seg = hold1;
         strcpy(symbols[num_syms++].sym,&globalstr[ptr]);
         return(TRUE);
         }
      }
   return(FALSE);
   }

int numval(str,val)  /* change hex value in str to a number */
                     /* return FALSE if invalid hex number, else TRUE */
   unsigned char *str;
   unsigned int *val;  /* Beware of passing an INT pointer to this routine
                           if the machine does not use low-byte first! */
   {
   int i,valid;
   unsigned int ch;

   i = 0;
   *val = 0;
   valid = FALSE;
   ch = toupper(str[i++]);
   while (ch)
      {
      valid = TRUE;
      if (isdigit(ch))
         *val = (*val << 4) + ch - '0';  /* a number */
      else
         if ((ch >= 'A') && (ch <= 'F'))  /* a letter */
            *val = (*val << 4) + ch - 'A' + 10;
         else
            return(FALSE);
      ch = toupper(str[i++]);
      }
   return(valid);
   }

int decval(str,val)  /* change dec value in str to a number */
                     /* return FALSE if invalid dec number, else TRUE */
   unsigned char *str;
   unsigned int *val;  /* Beware of passing an INT pointer to this routine
                           if the machine does not use low-byte first! */
   {
   int i,valid;
   unsigned int ch;

   i = 0;
   *val = 0;
   valid = FALSE;
   ch = str[i++];
   while (ch)
      {
      valid = TRUE;
      if (isdigit(ch))
         *val = (*val * 10) + ch - '0';  /* a number */
      else
         return(FALSE);
      ch = str[i++];
      }
   return(valid);
   }

parse_seg_and_ofs(seg,ofs,p,str)
   unsigned int *seg, *ofs, *p;
   unsigned char *str;
   {
   unsigned char temp[150];
   unsigned int i,ptr,hold1,hold2,flag,count;
   unsigned char ch,byte;

   ptr =
   i = 0;
   do
      {
      ch =
      temp[i] = str[ptr];
      i++;
      ptr++;
      } while ((ch) && (ch != ' ') && (ch != ':'));
   if (i < 6)  /* if this could be a valid symbol */
      {
      hold1 = parse_seg;  /* get our chosen segment */
      temp[i-1] = '\0';  /* back to the terminating char */
      if (ch == ':')  /* if a segment */
         {
         strupr(temp);
         flag = FALSE;
         if (strcmp(temp,"CS") == 0)
           {
           hold1 = reg_cs;
           flag = TRUE;
           }
         if ((strcmp(temp,"DS") == 0) || (strcmp(temp,"ES") == 0)
                                               || (strcmp(temp,"SS") == 0))
            {
            flag = TRUE;
            hold1 = reg_ss;
            }
         if (!flag)
            {
            if (numval(temp,&hold1) == FALSE)
               return(FALSE);  /* this is no good */
            }
         i = 0;
         do
            {
            ch =
            temp[i] = str[ptr];
            i++;
            ptr++;
            } while ((ch) && (ch != ' '));
         temp[i-1] = '\0';  /* back to the terminating char */
         if (i > 5)
            {
            return(FALSE);
            }
         }
      if (numval(temp,&hold2))
         {
         *ofs = hold2;
         *seg = hold1;
         *p = ptr;  /* say how far we got */
         return(TRUE);  /* and this is a good parse */
         }
      }
   return(FALSE);
   }

int read_symbol_file(fname)
   char *fname;
   {
   char temp[10];
   int filein;
   unsigned int skipped;
   unsigned char ch,byte;

   skipped =
   num_syms = 0;  /* start off clean */
   if ((filein = fopen(fname,"r")) == 0)
      {
      puts("Cannot open ");
      puts(fname);
      puts(" for reading.\nADDCOM aborted.\n");
      return(FALSE);
      }
   while ((num_syms < SYM_TABLE_SIZE) && (fgets(globalstr,148,filein)))
      {
      if (!parse_symbol())
         {
         if (!skipped)
            puts("\nInvalid symbols:\n\n");
         puts(globalstr);
         skipped++;
         }
      }
   fclose(filein);
   if (skipped)
      {
      itoa(skipped,temp);
      puts(temp);
      puts(" invalid symbols skipped.\n");
      }
   if (!num_syms)
      {
      puts("No symbols to load!\nADDCOM aborted.\n");
      return(FALSE);
      }
   if (num_syms == SYM_TABLE_SIZE)
      puts("Symbol table full.\n");
   qsort(symbols,num_syms,sizeof(struct symbol_table),compare_sym_addr);
   return(TRUE);
   }

int lookup_sym(seg,ofs)
   unsigned int seg,ofs;
   {
   int min, max, middle, i;
   struct symbol_table s;

   min = 0;
   max = num_syms - 1;
   s.seg = seg;
   s.ofs = ofs;
   do
      {
      middle = (min + max) / 2;
      if ((i = compare_sym_addr(&symbols[middle],&s)) == 0)
         return(middle);
      if (i > 0)
         max = middle - 1;
      else
         min = middle + 1;
      } while (min <= max);
   return(-1);
   }

int get_sym(name)  /* Symbols are NOT sorted by name! */
   char *name;
   {
   int i;

   for (i = 0; i < num_syms; i++)
      if (strcmp(symbols[i].sym,name) == 0)
         return(i);
   return(-1);
   }

read_header(filename)
   char *filename;
   {
   struct imageheader
      {
      unsigned int exesignature;
      unsigned int lenmod512;
      unsigned int sizeinpages;
      unsigned int numofrelocitems;
      unsigned int headersize;
      unsigned int minmem;
      unsigned int maxmem;
      unsigned int ssdisp;
      unsigned int spvalue;
      unsigned int checksum;
      unsigned int ipvalue;
      unsigned int csdisp;
      unsigned int relocdisp;
      unsigned int ovlynumber;
      } header;
   unsigned int count,len,flag;

   len = strlen(filename);
   flag = -1;  /* look for a period */
   for (count = len-1; count > 0; count--)
      if (filename[count] == '.')
         flag = count;   /* remember where it is */
   if ((flag > 0) && ((len-flag-1) == 3))  /* if a period was found
                                 and the extension length is 3 */
      {
      if ((filename[flag+1] == 'E') && (filename[flag+2] == 'X')
                                          && (filename[flag+3] == 'E'))
         {
         read(executable_file,&header,sizeof(struct imageheader));
                                                  /* read initial header */
         if (header.exesignature != 0x5a4d)
            puts("invalid .EXE file.\n");
         else
            {
            reg_cs = header.csdisp;  /* set registers */
            reg_ss = header.ssdisp;
            reg_ip = header.ipvalue;
            if (reg_cs != reg_ss)
               is_exefile = TRUE;
            }
         }
      }
   }

int iscomment(line)
   char *line;
   {
   while ((*line) && (*line != '\015') && (*line != '\012')
                  && ((*line == ' ') || (*line == '\t')))
      line++;
   if ((*line == ';') || (*line == '\015') || (*line == '\012'))
                 /* if blank or first non-blank is a semi-colon */
      return(TRUE);
   return(FALSE);
   }

int isdataline(line)
   char *line;
   {
   int temp;
   char templabel[LABEL_LEN], ch, ch2;

   if (iscomment(line))  /* let a comment be a data line, so that the
                            isdataline() loop does not end prematurely */
      return(TRUE);
   temp = 0;
   if (getlabel(line,templabel))  /* if a label, skip it */
      {
      temp = strlen(templabel);
      if (line[temp] == ':')
         temp++;
      }
   while (((ch=line[temp]) == ' ') || (ch == '\t'))
      temp++;
   if ((toupper(line[temp]) == 'D') && (isalpha(line[temp+1]))
           && ((line[temp+2] == ' ') || (line[temp+2] == '\t')))
      return(TRUE);
   return(FALSE);
   }

int iscodegenlabel(label)
   char *label;
   {
   int temp,temp2;

   if (((temp = strlen(label)) == 6)
           && (((is_exefile) && ((toupper(label[0]) == 'C')
                                               || (toupper(label[0]) == 'D')))
               || ((!is_exefile) && (toupper(label[0]) == 'H')))
           && (label[1] == '_')
           && (numval(&label[2],&temp2) == TRUE))
      return(TRUE);
   if ((temp == 5) && (!is_exefile) && (toupper(label[0]) == 'H')
                                     && (numval(&label[1],&temp2) == TRUE))
      return(TRUE);
   return(FALSE);
   }

int getlabel(line,label)  /* max label length is LABEL_LEN */
   char *line, *label;
   {
   int i;
   char ch,ch2;  /* use ch2 to keep stack aligned, just in case */

   for (i = 0; i < LABEL_LEN; i++)
      {
      ch = line[i];
      if ((ch == ' ') || (ch == '\t') || (ch == ':') || (ch == '\015')
                            || (ch == ';') || (ch == '\012') || (ch == '\0'))
         break;
      label[i] = ch;
      }
   if ((i == 0) || (i == LABEL_LEN))  /* if no label */
      return(FALSE);
   label[i] = '\0';  /* terminate the string */
   return(TRUE);
   }

replace_datasym(line)
   char *line;
   {
   int whichsym,temp,temp2;
   unsigned int ofs, seg;
   char label[LABEL_LEN], ch, ch2;

   if ((iscomment(line)) || (label_value(line,label,&ofs,&seg,&whichsym) == FALSE)
             || (whichsym != -1))  /* no label or is already a symbol */
      return;  /* no label or is already a sym, forget it */
   if ((whichsym = lookup_sym(seg,ofs)) == -1)
      return;
   temp = strlen(label);
   strcpy(newline,symbols[whichsym].sym);  /* use this as a temp string */
   temp2 = strlen(newline);  /* length of new label */
   if (line[temp] == ':')  /* if label ended in a colon */
      {
      strcat(newline,":");
      temp++;
      temp2++;
      }
   while ((line[temp] == ' ') || (line[temp] == '\t'))
      temp++;
   if (line[temp])    /* if more on line */
      {
      if ((line[temp] != '\015') && (line[temp] != '\012'))  /* if not EOL */
         {
         if (temp2 > 7)
            strcat(newline,nl_str);  /* separate onto two lines */
         strcat(newline,tab_str);  /* use uniform spacing */
         strcat(newline,&line[temp]);
         }
      else
         strcat(newline,nl_str);
      }
   else
      strcat(newline,nl_str);
   strcpy(line,newline);
   }

replace_labels(line)
   char *line;
   {
   int whichsym,line_ptr,temp,changed;
   unsigned int ofs, seg;
   char label[LABEL_LEN], ch, ch2;

   changed = FALSE;
   line_ptr = 0;
   if (getlabel(line,label))  /* if we have a label */
      {
      line_ptr = strlen(label);  /* then skip it */
      if (line[line_ptr] == ':')
         line_ptr++;
      }
   while ((line[line_ptr] == ' ') || (line[line_ptr] == '\t'))
      line_ptr++;             /* skip any spaces or tabs */
   line_ptr++;  /* skip the 'D' */
   if (toupper(line[line_ptr]) != 'W')  /* if could not be a label */
      return;
   line_ptr++;  /* skip the 'W' */
   while ((line[line_ptr] == ' ') || (line[line_ptr] == '\t'))
      line_ptr++;  /* skip out to the data */
   if (((ch = line[line_ptr]) == '\0') || (ch == '\015') || (ch == '\012'))
      return;  /* if EOL, we're done */
   while ((ch = line[line_ptr]) && (ch != ';') && (ch != '\015')
                    && (ch != '\012'))  /* until no data left */
      {
      ch = line[line_ptr];  /* initial conditions */
      while ((ch) && (!isalnum(ch)) && (ch != '\012') && (ch != '\015')
                        && (ch != ';'))
         ch = line[++line_ptr];  /* skip any garbage */
      temp = line_ptr;
      while ((isalnum(line[temp])) || (line[temp] == '_'))
         {
         templine[temp-line_ptr] = toupper(line[temp]);
         temp++;      /* copy label(?) into templine */
         }
      templine[temp-line_ptr] = '\0';
      if (temp-line_ptr == 0)  /* if blank, quit! */
         break;
      if (iscodegenlabel(templine))
         {
         if (templine[1] == '_')
            numval(&templine[2],&ofs);
         else
            numval(&templine[1],&ofs);
         if (templine[0] == 'D')  /* if a DATA symbol */
            seg = reg_ss;
         else
            seg = reg_cs;
         if ((whichsym = lookup_sym(seg,ofs)) != -1)
            {
            strcpy(templine2,&line[temp]);  /* save rest of line */
            strcpy(&line[line_ptr],symbols[whichsym].sym);  /* put in symbol */
            strcat(line,templine2);  /* re-attach rest of line */
            temp = line_ptr + strlen(symbols[whichsym].sym);
            changed = TRUE;
            }
         }
      line_ptr = temp;
      }
   if (!changed)
      return;   /* no need to play with comments */
   line_ptr = 0;  /* back to start of line */
   while ((ch = line[line_ptr]) && (ch != '\015') && (ch != '\012')
                                        && (ch != ';'))
      line_ptr++;
   if (line[line_ptr] == ';')  /* if we have a comment */
      {
      strcpy(templine,&line[line_ptr]);  /* save the comment */
      line_ptr--;  /* back past the ';' */
      while ((line[line_ptr] == ' ') || (line[line_ptr] == '\t'))
         line_ptr--;  /* back to end of data */
      line[line_ptr+1] = '\0';  /* stop after data */
      line_ptr =
      temp = 0;
      while (ch = line[line_ptr++])
         temp += (ch == '\t') ? (8 - temp%8) : 1;
      if ((temp + strlen(templine) > 79) || (temp >= tab_spacing))
         strcat(line,"  ");  /* just some spaces, already too long! */
      else
         {
         temp -= temp%8;  /* back to start of tab */
         changed = strlen(templine);  /* use changed as a temp */
         while (temp < tab_spacing)
            {
            strcat(line,tab_str);
            temp += 8;
            }
         }
      strcat(line,templine);  /* put comment back on line */
      }
   }

int label_value(line,label,ofs,seg,whichsym)  /* TRUE: ofs,seg = label val */
   char *line, *label;
   unsigned int *ofs, *seg, *whichsym;
   {
   unsigned int temp;

   if ((iscomment(line)) || (getlabel(line,label) == FALSE))
      return(FALSE);  /* no good, no label to use */
   if ((*whichsym = get_sym(label)) == -1)
      {
      if (label[1] == '_')
         temp = numval(&label[2],ofs);
      else
         temp = numval(&label[1],ofs);
      if (temp == FALSE)      /* bad label */
         return(FALSE);
      if (toupper(label[0]) == 'D')
         *seg = reg_ss;
      else
         *seg = reg_cs;
      }
   else
      {
      *ofs = symbols[*whichsym].ofs;
      *seg = symbols[*whichsym].seg;
      }
   return(TRUE);
   }

int find_file_pos(src,target)  /* return readflag,
                                  len of passed strs is LINE_LEN */
   char *src, *target;
   {
   unsigned int srcofs, srcseg, targetofs, targetseg, temp;
   char srclabel[LABEL_LEN], targetlabel[LABEL_LEN];
   long filepos, fseek();

   filepos = fseek(codegen_file,0L,1);  /* seek to current loc+0 == get loc */
   if (getlabel(src,srclabel))
      {
      if (label_value(src,srclabel,&srcofs,&srcseg,&temp))
         {
         targetofs = srcofs - 1;  /* so they are NOT equal to start with */
         do
            {
            if (fgets(target,LINE_LEN,codegen_file) == 0)
               {
               fseek(codegen_file,filepos,0);  /* go back to original pos */
               return(FALSE);  /* error on read, possibly EOF */
               }
            if (label_value(target,targetlabel,&targetofs,&targetseg,&temp) == FALSE)
              targetofs = srcofs - 1;  /* so they are NOT equal by accident! */
             } while ((targetofs != srcofs) || (targetseg != srcseg));
         return(TRUE);  /* if we got to here, there are no errors */
         }
      do
         {
         if (fgets(target,LINE_LEN,codegen_file) == 0)
            {
            fseek(codegen_file,filepos,0);  /* go back to original pos */
            return(FALSE);  /* error on read, possibly EOF */
            }
         if (getlabel(target,targetlabel) == FALSE)
            targetlabel[0] = '\0';  /* no leftover garbage */
         } while (strcmp(srclabel,targetlabel));
      return(TRUE);  /* to get here, there are no errors */
      }
   do
      {
      if (fgets(target,LINE_LEN,codegen_file) == 0)
         {
         fseek(codegen_file,filepos,0);  /* go back to original pos */
         return(FALSE);  /* error on read, possibly EOF */
         }
      } while (strcmp(src,target));  /* last ditch possibility */
   return(TRUE);  /* to get here, there are no errors */
   }

int check_for_offset(str)      /* returns loc of start of offset label in str,
                                  or 0 if not found */
   char *str;
   {
   int strptr, temp;
   char label[LABEL_LEN],ch, ch2;

   strptr = 0;
   while ((ch = str[strptr]) && (ch != ',') && (ch != ';') && (ch != '\015')
                  && (ch != '\012'))
      strptr++;
   if (ch != ',')
      return(0);   /* an offset must be in the second argument.  If there is
                      no second arguement, we're done */
   strptr++;  /* skip the comma */
   while ((ch = toupper(str[strptr])) && ((ch == ' ') || (ch == '\t')))
      strptr++;
   if (ch != 'O')
      return(0);  /* can't be OFFSET! */
   if (!getlabel(&str[strptr],label))
      return(0);  /* can't get OFFSET */
   for (temp = 0; label[temp]; temp++)
      label[temp] = toupper(label[temp]);  /* force upper case */
   if (strcmp(label,"OFFSET"))
      return(0);  /* isn't OFFSET */
   strptr += 6;
   while ((ch = str[strptr]) && ((ch == ' ') || (ch == '\t')))
      strptr++;
   return(strptr);  /* return the location of the label */
   }

int merge_lines(src,target)  /* move code from src to target, leave comments
                                in target alone.  Neither target nor src are a
                                comment line.  Return errflag. */
   char *src, *target;
   {
   int srcsym, targetsym, has_targetoffset;
   unsigned int srcofs, srcseg, targetofs, targetseg, temp, done;
   unsigned int has_srclabel, has_targetlabel, have_srcval, have_targetval;
   unsigned int targetlen, srclen;
   char srclabel[LABEL_LEN], targetlabel[LABEL_LEN], ch, ch2;

   srcsym =
   targetsym = -1;   /* initalize to not there */
   has_srclabel = getlabel(src,srclabel);
   has_targetlabel = getlabel(target,targetlabel);

   if (has_targetlabel)  /* if label on the target */
      {
      done =
      have_targetval = FALSE;  /* default: no value yet for target labels */
      do
         {
         if (!have_targetval)
            {
            have_targetval = label_value(target,targetlabel,&targetofs,
                                                       &targetseg,&targetsym);
                          /* I know I can parse the label again */
            if ((have_targetval) && (has_srclabel) /* if just found a value */
                       && (label_value(src,srclabel,&srcofs,&srcseg,&srcsym)))
               {
               if ((targetofs != srcofs) || (targetseg != srcseg))
                  return(ERR_NOMATCH_LABEL);  /* Oops!  non-matching labels! */
               }
            }
         temp = strlen(targetlabel);
         temp += (target[temp] == ':') ? 1 : 0;  /* if ends in a ':',
                                                    include in label length */
         if (!iscomment(&target[temp]))
            done = TRUE;
         else
            {
            if ((iscodegenlabel(targetlabel)) && (srcsym != -1))
               {
               strcpy(newline,symbols[srcsym].sym);  /* use as a temp string */
               strcat(newline,&target[strlen(targetlabel)]);
               strcpy(target,newline);  /* replace label with symbol */
               }
            if (fputs(target,output_file) == -1)
               return(ERR_WRITE_ERROR);  /* output error, time to quit */
            if (fgets(target,LINE_LEN,commented_file) == 0)
               return(ERR_BAD_READ_COM);  /* if can't read next line, we're doomed */
            if ((has_targetlabel = getlabel(target,targetlabel)) == FALSE)
               {
               targetlabel[0] = '\0';  /* make sure its clean! */
               if (!iscomment(target))
                  done = TRUE;
               }
            }
         } while (!done);
      }
   newline[0] = '\0';  /* use as a temp string, starting blank */
   targetlen = 0;
   if (has_targetlabel)
      {
      if ((has_srclabel) && (srcsym != -1))  /* if a source label for replacement */
         strcat(newline,symbols[srcsym].sym);
      else
         strcat(newline,targetlabel);
      targetlen = strlen(targetlabel);
      if (target[targetlen] == ':')
         {
         strcat(newline,":");
         targetlen++;
         }
      temp = targetlen;
      while ((ch=target[temp]) && ((ch == ' ') || (ch == '\t')))
         temp++;
      if ((ch != '\015') && (ch != '\012'))
         strcat(newline,tab_str);
      else
         strcat(newline,nl_str);
      }
   else
      strcat(newline,tab_str);
   while ((target[targetlen] == ' ') || (target[targetlen] == '\t'))
      targetlen++;
   temp = targetlen;
   while (isalnum(target[temp]))
      {
      templine[temp-targetlen] = toupper(target[temp]);
      temp++;   /* avoid side-affects */
      }
   templine[temp-targetlen] = '\0';
   if ((strlen(templine) == 2) && (templine[0] == 'D'))  /* if DATA */
      {
      replace_labels(target);
      strcat(newline,&target[targetlen]);
      strcpy(target,newline);  /* put final form back into target */
      return(ERR_NONE);  /* and that's all for here! */
      }
   srclen = 0;
   if (has_srclabel)
      {
      done = FALSE;
      do
         {
         srclen = strlen(srclabel);
         if (src[srclen] == ':')
            srclen++;
         if (iscomment(&src[srclen]))  /* if label only on this line */
            {
            if (fgets(src,LINE_LEN,codegen_file) == 0)
               return(ERR_BAD_READ_CG);  /* no read, we're done */
            srclen = 0;
            if (((has_srclabel = getlabel(src,srclabel)) == FALSE)
                       && (!iscomment(src)))
               done = TRUE;
            }
         else
            done = TRUE;
         } while (!done);
      }
   while ((src[srclen] == ' ') || (src[srclen] == '\t'))
      srclen++;
   temp = srclen;
   while (isalnum(src[temp]))
      {
      templine2[temp-srclen] = toupper(src[temp]);
      temp++;   /* avoid side-affects */
      }
   templine2[temp-srclen] = '\0';
   if (strcmp(templine,templine2))  /* if instructions not the same */
         return(ERR_NOMATCH_INSTR);   /* wrong instructions, and not DB/DW */
   has_targetoffset = check_for_offset(&target[targetlen]);
   temp = check_for_offset(&src[srclen]);
   if (((has_targetoffset) && (!temp)) || ((!has_targetoffset) && (temp)))
      {
      temp = targetlen + has_targetoffset;
      if ((has_targetoffset)
           && (label_value(&target[temp],targetlabel,&targetofs,&targetseg,&targetsym)))
         {
         if ((targetsym == -1) && ((targetsym = lookup_sym(targetseg,targetofs)) != -1))
            {
            ch = target[temp];
            target[temp] = '\0';  /* terminate string at the label */
            strcat(newline,&target[targetlen]);
            target[temp] = ch;  /* put back 1st char of label */
            strcat(newline,symbols[targetsym].sym);  /* add new symbol */
            targetlen = temp + strlen(targetlabel);  /* add length of label */
            while ((ch = target[targetlen]) && (ch != '\015') && (ch != '\012')
                                               && (ch != ';'))
               targetlen++;
            if (ch == ';')    /* if a comment is present */
               {
               temp =
               done = 0;  /* use as a counter */
               while (newline[temp])
                  {
                  done += (newline[temp] == '\t') ? (8 - done%8) : 1;
                  temp++;
                  }
               if ((done >= tab_spacing) || (done+strlen(&target[targetlen]) > 79))
                  strcat(newline,"  ");
               else
                  {
                  do
                     {
                     strcat(newline,tab_str);
                     done += 8-done%8;
                     } while (done < tab_spacing);
                  }
                  strcat(newline,&target[targetlen]);  /* add comment */
               }
            else
               strcat(newline,nl_str);  /* make sure newline has a new line! */
            }
         else
            strcat(newline,&target[targetlen]);
         }
      else
         strcat(newline,&target[targetlen]);
      }
   else
      {
      temp = srclen;     /* get start of replacement instruction */
      while ((ch = src[temp]) && (ch != '\015') && (ch != '\012') && (ch != ';'))
         temp++;  /* find end of instruction */
      while ((temp)
              && (((ch = src[temp]) == ';') || (ch == '\015')
                       || (ch == '\012') || (ch == ' ') || (ch == '\t')))
         temp--;  /* back up to real end of instruction */
      temp++;  /* temp pointed to the last char in the instr, skip past it */
      src[temp] = '\0';  /* terminate string there */
      strcat(newline,&src[srclen]);  /* put instruction in new line */
      while ((ch = target[targetlen]) && (ch != '\015') && (ch != '\012')
                                               && (ch != ';'))
         targetlen++;
      if (ch == ';')    /* if a comment is present */
         {
         temp =
         done = 0;  /* use as a counter */
         while (newline[temp])
            {
            done += (newline[temp] == '\t') ? (8 - done%8) : 1;
            temp++;
            }
         if ((done >= tab_spacing) || (done+strlen(&target[targetlen]) > 79))
            strcat(newline,"  ");
         else
            {
            do
               {
               strcat(newline,tab_str);
               done += 8-done%8;
               } while (done < tab_spacing);
            }
         strcat(newline,&target[targetlen]);  /* add comment */
         }
      else
         strcat(newline,nl_str);  /* make sure newline has a new line! */
      }
   strcpy(target,newline);  /* put final form back into target */
   return(ERR_NONE);
   }

merge_files()
   {
   int errflag,readflag,done,count,hexval,sync_flag;
   long fseek(), cg_filepos;

   errflag = ERR_NONE;  /* no errors */
   com_line[0] = '\0';  /* initally blank */

   done = FALSE;
   do
      {
      if (fgets(com_line,LINE_LEN,commented_file) == 0)
         errflag = ERR_EOF;
      else
         {
         if (iscomment(com_line))
            {
            if (fputs(com_line,output_file) == -1)
               errflag = ERR_WRITE_ERROR;  /* error, we're done */
            }
         else
            {
            sync_flag = FALSE;  /* files not in sync */
            if (isdataline(com_line))
               {
               do
                  {
                  if (!iscomment(com_line))
                     replace_labels(com_line);
/*
   Note:  must do replace_labels() first, or else the newline inserted after
          a long label will hide the DW stuff!
*/
                  replace_datasym(com_line);
                  if (fputs(com_line,output_file) == -1)
                     errflag = ERR_WRITE_ERROR;
                  if (fgets(com_line,LINE_LEN,commented_file) == 0)
                     errflag = ERR_EOF;
                  } while ((!errflag) && (isdataline(com_line)));
               if (find_file_pos(com_line,cg_line) == FALSE)
                           /* I have no idea what this line is */
                  {
                  do       /* copy lines until sync or error */
                     {
                     if (fputs(com_line,output_file) == -1)
                        errflag = ERR_WRITE_ERROR;
                     if (fgets(com_line,LINE_LEN,commented_file) == 0)
                        errflag = ERR_EOF;
                     } while ((!errflag)
                            && (find_file_pos(com_line,cg_line) == FALSE));
                  }
               sync_flag = TRUE;  /* files are in sync */
               }
            if (!errflag)
               {
               do
                  {
                  if ((!sync_flag) && ((fgets(cg_line,LINE_LEN,codegen_file) == 0)))
                     {
                     errflag = ERR_CG_FILE_EOF;
                     do
                        {
                        if (fputs(com_line,output_file) == -1)
                           errflag = ERR_WRITE_ERROR;
                        } while ((errflag == ERR_CG_FILE_EOF)
                             && (fgets(com_line,LINE_LEN,commented_file) != 0));
                     }
                  else
                     {
                     sync_flag = FALSE;  /* need to read it next time */
                     if (!iscomment(cg_line))
                        {
                        cg_filepos = fseek(codegen_file,0L,1);
                                   /* seek to current loc+0 == get loc */
                        if ((errflag = merge_lines(cg_line,com_line))
                                       == ERR_NOMATCH_LABEL)
                              /* merge cg into com */
                              /* can have error in mismatched symbols */
                           {
                           errflag = ERR_NONE;  /* reset error */
                           fseek(codegen_file,cg_filepos,0);  /* reset ptr */
                           if (fputs(com_line,output_file) == -1)
                              errflag = ERR_WRITE_ERROR;
                           else
                              {
                              if (fgets(com_line,LINE_LEN,commented_file) == 0)
                                 errflag = ERR_EOF;
                              else
                                 {
                                 if (find_file_pos(com_line,cg_line) == FALSE)
                                        /* I have no idea what this line is */
                                    {
                                    do       /* copy lines until sync or error */
                                       {
                                       if (fputs(com_line,output_file) == -1)
                                          errflag = ERR_WRITE_ERROR;
                                       if (fgets(com_line,LINE_LEN,commented_file) == 0)
                                          errflag = ERR_EOF;
                                       } while ((!errflag)
                                                  && (find_file_pos(com_line,cg_line) == FALSE));
                                    }
                                 sync_flag = TRUE;  /* files are in sync */
                                 }
                              }
                           }
                        else
                           if (fputs(com_line,output_file) == -1)
                              errflag = ERR_WRITE_ERROR;  /* error, we're done */
                        }
                     }
                  } while ((iscomment(cg_line)) && (!errflag));
               }  /* if errflag after isdataline */
            }  /* else of ifcomment on comline */
         }  /* else of if fgets on com_line ... */
      } while ((!errflag) && (!done));
   switch (errflag)
      {
      case ERR_NONE:
      case ERR_EOF:
         break;
      case ERR_NOMATCH_INSTR:
         puts("Instructions in commented_file and codegen_file not equivalent.\n");
         break;
      case ERR_BAD_READ_COM:
         puts("Read error (or EOF) at unrecoverable location in commented_file.\n");
         break;
      case ERR_BAD_READ_CG:
         puts("Read error (or EOF) at unrecoverable location in codegen_file.\n");
         break;
      case ERR_WRITE_ERROR:
         puts("Error writing to output_file.  Check for disk full or write protect.\n");
         break;
      case ERR_CG_FILE_EOF:
         puts("Unexptected error (or EOF) in codegen_file.\n");
         puts("Remainder of commented_file copied to output_file.\n");
         break;
      }
   if ((errflag != ERR_NONE) && (errflag != ERR_EOF)
                                           && (errflag != ERR_CG_FILE_EOF))
      puts("\nADDCOM aborted.\n");
   }

main(argc,argv)
   int argc;
   char *argv[];
   {
   unsigned int temp;

   puts("ADDCOM V1.0   copyright (c) 1988 by William Couture.\n\n");
   is_exefile = FALSE;     /* default to .COM file */

   reg_cs =
   reg_ss = 0;        /* make sure segs start at 0 */
   reg_ip = 0x100;  /* let programs think they have a psp */

   if (argc < 6)
      {
      puts("Error, useage is:\n");
      puts("ADDCOM commented_file codegen_file output_file symbol_file\n");
      puts("                                        executable_file [comment_spacing]\n");
      exit(2);
      }
   if ((commented_file = fopen(argv[1],"r")) == 0)
      {
      puts("Cannot open ");
      puts(argv[1]);
      puts(" for reading.\nADDCOM aborted.\n");
      exit(2);
      }
   if ((codegen_file = fopen(argv[2],"r")) == 0)
      {
      puts("Cannot open ");
      puts(argv[2]);
      puts(" for reading.\nADDCOM aborted.\n");
      exit(2);
      }
   if ((output_file = open(argv[3],MODE_READ)) != -1)
      {
      close(output_file);
      puts("The output file ");
      puts(argv[3]);
      puts(" already exists.\n");
      puts("Do you wish to over-write it? (Y/N, return = No) ");
      gets(globalstr);
      if (toupper(globalstr[0]) != 'Y')
         {
         puts("ADDCOM aborted.\n");
         exit(2);
         }
      }
   if ((output_file = fopen(argv[3],"w")) == 0)
      {
      puts("The file ");
      puts(argv[3]);
      puts("is read only.\nADDCOM aborted.\n");
      exit(2);
      }
   if ((executable_file = open(argv[5],MODE_READ)) == -1)
      {
      puts("Cannot open ");
      puts(argv[5]);
      puts(" for reading.\nADDCOM aborted.\n");
      exit(2);
      }
   read_header(argv[5]);    /* determine segmentation */
   close(executable_file);  /* don't need this file anymore */
   if ((read_symbol_file(argv[4]) == FALSE))
      exit(2);
   if (argc > 6)
      {
      if ((!decval(argv[6],&temp)) || (temp < 20) || (temp > 78))
         puts("Illegal comment spacing--value ignored.\n");
      else
         tab_spacing = temp;
      }
   merge_files();  /* and go have "fun"! */
   fclose(commented_file);   /* close and update all files */
   fclose(codegen_file);
   fclose(output_file);
   }
