/*
   This file is copyright (c) 1986 by William Couture.  All rights reserved.

   Modified 4/88 for CODEGEN.   Do not move back to VIM as is!
*/


/*
   The instruction decoder, string generating version.  This program
   contains the routines to look up instructions in the instruction
   matrix (optables) and determine the corresponding instruction.

   Also, it will recognize addressing modes and segment prefixes to
   determine the byte or word register or memory addressed, and
   load the values for both source and destination.  For instructions
   like the string primitives, or XLAT, etc., where the addressing
   mode is not obvious, the memory addressing is done from the instruction
   simulation itself.

   This routine is used for three purposes:
      1) The Trace command, where the instructions are displayed as
         they are executed.
      2) The Reg command, to display both registers and the current
         instruction.
      3) The Unassemble command, to show the source code to a program.

   For the execution of the Go command, the routine decodens() is used.
   This routine is formed by removing ALL string references from this
   routine (the equivalent to the modename() routine is the eabyte()
   routine).  This is done to speed up execution, as the strings would
   not be used anyway.  Note that comments in decode() and decodens()
   are not always the same.  Check them both out for information.
*/

/*
   Note:  The usual check for BYTE/WORD instruction is:
      if (instruction.flags & BYTE)
   The lack of a BYTE flag is taken to mean a word flag.
*/

#include <compiler.h>
#include <codegen.h>
#include <opdefs.h>
#include <regstruc.h>

#define VALID 1  /* instruction decoded was valid */
#define INVALID 0  /* instruction decoded was invalid */

extern struct optable firstbyte[256];  /* from optables.c */
extern struct optable secondbyte[13][8];  /* from optables.c */
extern struct optable pmbyte[7];        /* from optables.c */
extern struct optable instruction;  /* from VIM.C */

extern struct tofrom dest;  /* from VIM.C */
extern struct tofrom source;  /* from VIM.C */

extern union regunion regs;  /* from VIM.C */

extern unsigned int prefix;  /* from VIM.C */
extern unsigned int this_prefix;
extern unsigned int prefixip;  /* used to hold address of executing instruction
                                  and the LAST prefix in case of interrupt
                                  during string primitive */
 
extern int mod,reg,rm;  /* from VIM.C */

extern unsigned int startip;  /* IP of this this instruction, including
                                prefixes */

extern unsigned int show_ptrs;

extern unsigned char byte_one;

extern char opnames[113][7];  /* from OPNAMES.C */
extern char regnames[23][3];  /* from OPNAMES.C */

unsigned int instr_ofs,instr_seg,holdreg;

extern unsigned int mixbytereg[8];  /* defined in VIM.C */

extern char tab[2];

extern unsigned int fetchbyte(), fetchword();

char hexdig(ch)  /* return the ASCII char corresponding to the HEX value */
   char ch;
   {
   return(ch > '\011' ? (ch-'\012' + 'A') : (ch+'0'));
   }

char *hexbyte(src,target)  /* encode the first two bytes of src into target
                              as ascii, and return the address of target.
                              Note that this is sometimes called with the
                              form &intvar for src.  THIS ONLY WORKS ON A
                              MACHINE THAT USES LOW BYTE FIRST!!
                              Sorry about that if you want to port it, but
                              I was lazy. */
   char *src, *target;
   {
   target[0] = hexdig( (src[0] >> 4) & 0x0f);
   target[1] = hexdig(src[0] & 0x0f);
   target[2] = '\0';
   return(target);
   }

char *hexword(src,target)  /* As for hexbyte, but for a two byte word.
                              The comment about being called with
                              &intvar for src applies DOUBLY here! */
   char *src, *target;
   {
   target[0] = hexdig( (src[1] >> 4) & 0x0f);
   target[1] = hexdig(src[1] & 0x0f);
   target[2] = hexdig( (src[0] >> 4) & 0x0f);
   target[3] = hexdig(src[0] & 0x0f);
   target[4] = '\0';
   return(target);
   }

char *hexlong(src,target)  /* As for hexbyte, but for 4 byte words */
   char *src, *target;
   {
   target[0] = hexdig( (src[3] >> 4) & 0x0f);
   target[1] = hexdig(src[3] & 0x0f);
   target[2] = hexdig( (src[2] >> 4) & 0x0f);
   target[3] = hexdig(src[2] & 0x0f);
   target[4] = ':';   /* longs are usually in the form xxxx:yyyy */
   target[5] = hexdig( (src[1] >> 4) & 0x0f);
   target[6] = hexdig(src[1] & 0x0f);
   target[7] = hexdig( (src[0] >> 4) & 0x0f);
   target[8] = hexdig(src[0] & 0x0f);
   target[9] = '\0';
   return(target);
   }

char *addoff8(src,ip,target)  /* return a string that is the sum of an
                                 8 bit signed displacement and the
                                 instruction pointer.  The address
                                 of the passed target string is returned */
   char *src, *target;
   int ip;
   {
   char ch[4], *hexword();  /* let C know the return type */

   if (src[0] & SIGNEDB)
      ip -= -(src[0] & 0x7f) + 128;  /* sign extend byte when working with
                                        unsigned vars */
   else
      ip += src[0];
   ch[0] = ip & 0xff;
   ch[1] = (ip >> 8) & 0xff;
   return(hexword(ch,target));
   }

char *addoff16(src,ip,target)  /* return a string corresponding to the sum
                                  of a signed 16 bit displacement and the
                                  instruction pointer.  The address of the
                                  target string is returned. */
   char *src, *target;
   int ip;
   {
   char ch[4], *hexword();  /* let C know the return type */
   int i;

   i = src[0] + (src[1] << 8);
   ip += i;
   ch[0] = ip & 0xff;
   ch[1] = (ip >> 8) & 0xff;
   return(hexword(ch,target));
   }

char *modename(hexout,hold)  /* perform EA byte addressing, and
                                           also return a string containing
                                           the mode name.  The address of
                                           the return string is the value
                                           of the function. */
   char *hexout, *hold;
   {
   long i;
   unsigned char filler,byte;
   unsigned char ch[4];
   char temp[80],*hexbyte(),*hexword();

   strcpy(hold,"");
   if (mod < 3)  /* if memory addressing */
      {
      switch(rm)  /* get addressing mode */
         {
         case 0:
            strcat(hold,"[BX+SI");
            instr_ofs = regs.r.bx.word + regs.r.si;
            break;
         case 1:
            strcat(hold,"[BX+DI");
            instr_ofs = regs.r.bx.word + regs.r.di;
            break;
         case 2:
            strcat(hold,"[BP+SI");
            instr_ofs = regs.r.bp + regs.r.si;
            instr_seg = regs.r.ss; /* BP register defaults to stack segment */
            break;
         case 3:
            strcat(hold,"[BP+DI");
            instr_ofs = regs.r.bp + regs.r.di;
            instr_seg = regs.r.ss;
            break;
         case 4:
            strcat(hold,"[SI");
            instr_ofs = regs.r.si;
            break;
         case 5:
            strcat(hold,"[DI");
            instr_ofs = regs.r.di;
            break;
         case 6:
            if (mod)
               {
               strcat(hold,"[BP");
               instr_ofs = regs.r.bp;
               instr_seg = regs.r.ss;
               }
            else
               {  /* mod=0 means direct 16 bit address */
               ch[0] = fromqueue();
               ch[1] = fromqueue();
               instr_ofs = (ch[1] << 8) + ch[0];
               strcat(hexout,hexbyte(ch,temp));
               strcat(hexout,hexbyte(&ch[1],temp));
               strcat(hold,"[");
               strcat(hold,hnumber(instr_ofs,temp,instr_seg,regs.r.cs,startip));
               }
            break;
         case 7:
            strcat(hold,"[BX");
            instr_ofs = regs.r.bx.word;
            break;
         }
      switch (mod)
         {
         case 1:  /* add one byte displacement */
            byte = fromqueue();
            strcat(hexout,hexbyte(&byte,temp));
            if (byte > 0x7f)
               {
               instr_ofs += (-1 ^ 0xff) | byte;  /* that should be machine
                                              and compiler independent */
               strcat(hold,"-");
               ch[0] = -(byte & 0x7f) + 128;
               strcat(hold,numberh_byte(ch[0],temp));
               }
            else
               {
               instr_ofs += byte;
               strcat(hold,"+");
               strcat(hold,numberh_byte(byte,temp));
               }
            break;
         case 2:  /* add two byte displacement */
            ch[0] = fromqueue();
            ch[1] = fromqueue();
            strcat(hexout,hexbyte(ch,temp));
            strcat(hexout,hexbyte(&ch[1],temp));
            i = ch[0] + (ch[1] << 8);
            instr_ofs += i;  /* WATCH THIS ON A MACHINE THAT DOES NOT HAVE A
                              TWO BYTE INTEGER */
/*            if (i > 0x8000)
               {
               strcat(hold,"-");
               i = -(i & 0x7fff) + 32768L;  (* machine independent?? *)
               ch[0] = i & 0x00ff;
               ch[1] = (i >> 8) & 0x7f;
               strcat(hold,hnumber((unsigned int)i,temp,instr_seg,regs.r.cs.startip));
               }
            else
               { */
               strcat(hold,"+");
               if (prefix & SEG)
                  getseg();  /* codegen */
               strcat(hold,hnumber((unsigned int)i,temp,instr_seg,regs.r.cs,startip));
/*               } */
            break;
         }
      strcat(hold,"]");
      }
   else  /* register addressing */
      {
      if (instruction.flags & BYTE)
         {
         strcat(hold,regnames[rm+DAL-DAX]);
         holdreg = mixbytereg[rm];
         }
      else
         {
         strcat(hold,regnames[rm]);
         holdreg = rm;
         }
      }
   return(hold);
   }

void getseg()  /* Set seg for memory addressing to override value.  Called by
                  both decode() and decodens() */
   {
   switch (prefix & SEG)
      {
      case CSSEG:
         instr_seg = regs.r.cs;
         break;
      case DSSEG:
         instr_seg = regs.r.ds;
         break;
      case ESSEG:
         instr_seg = regs.r.es;
         break;
      case SSSEG:
         instr_seg = regs.r.ss;
         break;
      }
   }

void sourceswitch(hexout,out,ptrhold)
   char *hexout,*out,*ptrhold;
       /* NOTE:  This routine was originally part of the decode() routine,
          but that grew too big for the compiler, and so this piece was
          moved here.  Rather clumsy, but the best I can do.
        */
   {
   int i;
   unsigned char filler,byte;
   unsigned char hold[80];
   char temp[80],*hexbyte(),*hexword(),*hexlong();
   
   switch (instruction.source)
      {
      case S8:
         strcat(out,regnames[reg+SAL-SAX]);
         source.where = REGISTER;
         source.reg = mixbytereg[reg];
         source.value = (UNSIGNED_LONG)regs.bytearray[mixbytereg[reg]];
         break;
      case S16:
         strcat(out,regnames[reg]);
         source.where = REGISTER;
         source.reg = reg;
         source.value = (UNSIGNED_LONG)regs.wordarray[reg];
         break;
      case SDATA8:
         byte = fromqueue();
         strcat(hexout,hexbyte(&byte,temp));
         strcat(out,numberh_byte(byte,temp));
         source.value = (UNSIGNED_LONG)byte;
         break;
      case SDATA16:
         hold[0] = fromqueue();
         hold[1] = fromqueue();
         strcat(hexout,hexbyte(hold,temp));
         strcat(hexout,hexbyte(&hold[1],temp));
         strcat(out,hexword(hold,temp));
         source.value = (UNSIGNED_LONG)((hold[1] << 8) + hold[0]);
         break;
      case SXDATA8:
         byte = fromqueue();
         strcat(hexout,hexbyte(&byte,temp));
         if (byte > 0x7f)
            {
            strcat(out,"-");
            source.value = (-1 ^ 0xff) | byte;  /* that should be machine
                                              and compiler independent */
            hold[0] = -(byte & 0x7f) + 128;
            strcat(out,numberh_byte(hold[0],temp));
            }
         else
            {
            strcat(out,"+");
            source.value = (UNSIGNED_LONG)byte;
            strcat(out,numberh_byte(byte,temp));
            }
         break;
      case SOFF8:
         byte = fromqueue();
         strcat(hexout,hexbyte(&byte,temp));
         strcat(out,addoff8(&byte,regs.r.ip,temp));
         if (byte & SIGNEDB)
            source.value = (UNSIGNED_LONG)(regs.r.ip + (0xff00 + byte));
               /* note: this assumes 16 bit unsigned int */
         else
            source.value = (UNSIGNED_LONG)(regs.r.ip + byte);
         break;
      case SOFF16:
         hold[0] = fromqueue();
         hold[1] = fromqueue();
         strcat(hexout,hexbyte(hold,temp));
         strcat(hexout,hexbyte(&hold[1],temp));
         strcat(out,addoff16(hold,regs.r.ip,temp));
         source.value = (UNSIGNED_LONG)(regs.r.ip + ((hold[1] << 8) + hold[0]));
         break;
      case SADDR16:
         source.where = MEMORY;  /* codegen */
         hold[0] = fromqueue();
         hold[1] = fromqueue();
         strcat(hexout,hexbyte(hold,temp));
         strcat(hexout,hexbyte(&hold[1],temp));
         if (ptrhold[0])
            strcat(out,ptrhold);
         strcat(out,"[");
         if (prefix & SEG)
            getseg();
         strcat(out,hnumber((hold[1] << 8) + hold[0],temp,instr_seg,regs.r.cs,startip));
         strcat(out,"]");
         source.seg = instr_seg;
         source.ofs = (hold[1] << 8) + hold[0];
         if (instruction.flags & BYTE)
            source.value = (UNSIGNED_LONG)fetchbyte(instr_seg,source.ofs);
         else
            source.value = (UNSIGNED_LONG)fetchword(instr_seg,source.ofs);
         break;
      case SADDR32:
         for (i = 0; i < 4; i++)
            {
            hold[i] = fromqueue();
            strcat(hexout,hexbyte(&hold[i],temp));
            }
         strcat(out,hexlong(hold,temp));
         source.ofs = (hold[1] << 8) + hold[0];
         source.seg = (hold[3] << 8) + hold[2];
           /* CALLF and JMPF don't know if their address is coming
              from immediate data or a memory reference */
         source.value = (UNSIGNED_LONG)source.seg;
         source.value = (source.value << 16) + ((UNSIGNED_LONG)source.ofs);
         break;
      case SL:
         strcat(out,regnames[reg+SES-SAX]);
         source.where = SEGREG;
         source.reg = reg + 8;
         source.value = (UNSIGNED_LONG)regs.wordarray[source.reg];
         break;
      case SM:
         if (ptrhold[0])
            strcat(out,ptrhold);
         if (prefix & SEG)
            getseg();  /* codegen */
         strcat(out,modename(hexout,hold));
         if (prefix & SEG)
            getseg();
         if (instruction.flags & BYTE)
            {
            if (mod < 3)
               {
               source.where = MEMORY;  /* codegen */
               source.seg = instr_seg;
               source.ofs = instr_ofs;
               source.value = (UNSIGNED_LONG)fetchbyte(instr_seg,instr_ofs);
               }
            else
               {
               source.where = REGISTER;
               source.value = (UNSIGNED_LONG)regs.bytearray[holdreg];
               source.reg = holdreg;
               }
            }
         else
            {
            if (mod < 3)
               {
               source.where = MEMORY;  /* codegen */
               source.seg = instr_seg;
               source.ofs = instr_ofs;
               source.value = (UNSIGNED_LONG)fetchword(instr_seg,instr_ofs);
               if (instruction.flags & ADDR32)  /* for CALLF and JMPF */
                  {
                  if (instruction.source == SADDR32)
                     source.where = MEM_REG_UNKNOWN;  /* codegen */
                  source.value &= 0xffffL;
                  source.value += ((UNSIGNED_LONG)fetchword(instr_seg,instr_ofs+2)) << 16;
                  }
               }
            else
               {
               source.where = REGISTER;
               source.value = (UNSIGNED_LONG)regs.wordarray[holdreg];
               source.reg = holdreg;
               }
            }
         break;
      }
   source.oldvalue = source.value;
   }

int decode(hexout,out)
   char *hexout, *out;
   {
   int i;
   char temp[80],ptrhold[20],*hexbyte(),*hexword(),*hexlong();
   unsigned char filler,byte;
   unsigned char hold[80];
   
   mod = 0;  /* clear everything */
   reg = 0;
   rm = 0;
   hexout[0] = '\0';
   out[0] = '\0';
   ptrhold[0] = '\0';
   strcat(out,tab);
   byte_one = fromqueue();
   strcat(hexout,hexbyte(&byte_one,temp));
      /* Look up 1st byte in the instruction matrix.  The DeSmet compiler does
      not support structure assignment, so individual assignments are used. */
   instruction.name = firstbyte[byte_one].name;
   instruction.flags = firstbyte[byte_one].flags;
   instruction.dest = firstbyte[byte_one].dest;
   instruction.source = firstbyte[byte_one].source;
   instr_seg = regs.r.ds;  /* assume memrefs are relative to ds */
   if (instruction.flags & PREFIX)
      {
      prefixip = regs.r.ip-1;  /* ip already updated in fromqueue() */
      switch (instruction.name)
         {
         case SEGCS:
            prefix |= CSSEG;
            this_prefix = CSSEG;
            break;
         case SEGDS:
            prefix |= DSSEG;
            this_prefix = DSSEG;
            break;
         case SEGES:
            prefix |= ESSEG;
            this_prefix = ESSEG;
            break;
         case SEGSS:
            prefix |= SSSEG;
            this_prefix = SSSEG;
            break;
         case REPZ:
            prefix |= REPZERO;
            this_prefix = REPZERO;
            break;
         case REPNZ:
            prefix |= REPNOTZERO;
            this_prefix = REPNOTZERO;
            break;
         case LOCK:                  /* for CODEGEN */
            prefix |= LOCK_PREFIX;
            this_prefix = LOCK_PREFIX;
            break;
         }
       }
   if ((byte_one >= 0xa0) && (byte_one <= 0xa3) && (show_ptrs))  /* codegen */
      if (byte_one & 1)
         strcat(ptrhold,"WORD PTR ");
      else
         strcat(ptrhold,"BYTE PTR ");
   if (instruction.name == PMBYTE)  /* 286 protected instructions */
      {
      byte = fromqueue();
      strcat(hexout,hexbyte(&byte,temp));
      if (byte < 7)  /* if legal! */
         {
         instruction.name = pmbyte[byte].name;
         instruction.flags = pmbyte[byte].flags;
         instruction.source = pmbyte[byte].source;
         instruction.dest = pmbyte[byte].dest;
         }
      else
         instruction.name = UNDEF;  /* invalid instruction */
      }
   if ((instruction.dest < DDATA8) || (instruction.source < SDATA8))
      {  /* if EA byte present */
      byte = fromqueue();
      strcat(hexout,hexbyte(&byte,temp));
      mod = (byte >> 6) & 0x03;
      reg = (byte >> 3) & 0x07;
      rm = byte & 0x07;
      if (instruction.name == UNKNOWN)
/*         if ((instruction.dest == DK) && (mod == 3) && ((reg == 3) || (reg == 5)) )
                      {* JMP FAR AX, CALL FAR DI, etc, are illegal *}
            instruction.name = UNDEF;  handle weird decode cases
         else */
            {  /* decode second byte of instruction */
            instruction.name = secondbyte[instruction.dest - DA][reg].name;
            instruction.flags = secondbyte[instruction.dest - DA][reg].flags;
            instruction.source = secondbyte[instruction.dest - DA][reg].source;
            instruction.dest = secondbyte[instruction.dest - DA][reg].dest;
            }
      if (((instruction.dest == DL) || (instruction.source == SL)) && (reg > 3))
         instruction.name = UNDEF;  /* segreg name must be less than 3 */
      if ((mod == 3)
              && ((instruction.name == CALLF) || (instruction.name == JMPF)
                 || (instruction.name == BOUND) || (instruction.name == LES)
                 || (instruction.name == LDS) || (instruction.name == LEA)
                 || (instruction.name == SGDT) || (instruction.name == SIDT)
                 || (instruction.name == LGDT) || (instruction.name == LIDT)))
         instruction.name = UNDEF;
      if ((mod < 3) && (instruction.dest != DL) && (instruction.source != SL)
           && ( ( ((instruction.dest == DZ) || (instruction.dest < D8) || (show_ptrs)) && (instruction.source < S8))
           || ( (instruction.dest < D8) && ((instruction.source == SZ) || (instruction.source < S8) || (show_ptrs)) ) ) )
   /* show_ptrs is from CODEGEN */
         if (instruction.flags & BYTE)  /* byte or word memref? */
            strcat(ptrhold,"BYTE PTR ");
         else
            strcat(ptrhold,"WORD PTR ");
      }
   if (instruction.flags & AADAAM)  /* AAD and AAM are more weird cases */
      {
      byte = fromqueue();
      strcat(hexout,hexbyte(&byte,temp));
      if (byte != 0x0a)
         instruction.name = UNDEF;
      }
   if (instruction.name > DBYTE)  /* if valid instruction */
      {
      strcat(out,opnames[instruction.name]);
      strcat(out,tab);
      if ((instruction.dest >= DAX) && (instruction.dest <= DBH))
         {  /* if destination is a register */
         strcat(out,regnames[instruction.dest-DAX]);
         if (instruction.dest >= DAL)
            {
            dest.where = REGISTER;
            dest.reg = mixbytereg[instruction.dest - DAL];
            }
         else
            if (instruction.dest >= DFL)
               {
               dest.where = REGISTER;
               if (instruction.dest == DRAL)
                  dest.reg = mixbytereg[0];
               if (instruction.dest == DRAX)
                  dest.reg  = 0;
               }
            else
               if (instruction.dest >= DES)
                  {
                  dest.where = SEGREG;
                  dest.reg = instruction.dest - DES + 8;
                  }
               else
                  {
                  dest.where = REGISTER;
                  dest.reg = instruction.dest - DAX;
                  }
         if (instruction.flags & BYTE)
            dest.value = (UNSIGNED_LONG)regs.bytearray[dest.reg];
         else
            dest.value = (UNSIGNED_LONG)regs.wordarray[dest.reg];
         dest.oldvalue = dest.value;
         }
      else
         {
         switch (instruction.dest)
            {
            case D8:
               strcat(out,regnames[reg+DAL-DAX]);
               dest.where = REGISTER;
               dest.reg = mixbytereg[reg];
               dest.value = (UNSIGNED_LONG)regs.bytearray[dest.reg];
               break;
            case D16:
               strcat(out,regnames[reg]);
               dest.where = REGISTER;
               dest.reg = reg;
               dest.value = (UNSIGNED_LONG)regs.wordarray[dest.reg];
               break;
            case DDATA8:  /* used for the OUT instruction */
               byte = fromqueue();
               strcat(hexout,hexbyte(&byte,temp));
               strcat(out,temp);
               dest.value = (UNSIGNED_LONG)byte;
               break;
             case DDATA16:  /* used for the ENTER instruction */
               hold[0] = fromqueue();
               hold[1] = fromqueue();
               strcat(hexout,hexbyte(hold,temp));
               strcat(hexout,hexbyte(&hold[1],temp));
               strcat(out,numberh_word((hold[1] << 8) + hold[0],temp));
               dest.value = (UNSIGNED_LONG)((hold[1] << 8) + hold[0]);
               break;

/*          case DXDATA8:   copied for symmetry, these cases are not
                            used for destinations.  If they become
                            necessary, make sure they are current with
                            the source definitions

               byte = fromqueue();
               strcat(hexout,hexbyte(&byte,temp));
               if (byte > 0x7f)
                  {
                  strcat(out,"-");
                  hold[0] = -(byte & 0x7f) + 128;
                  strcat(out,hexbyte(hold,temp));
                  }
               else
                  {
                  strcat(out,"+");
                  strcat(out,hexbyte(&byte,temp));
                  }
               break;
            case DOFF8:
               byte = fromqueue();
               strcat(hexout,hexbyte(&byte,temp));
               strcat(out,addoff8(byte,regs.r.ip,temp));
               break;
            case DOFF16:
               hold[0] = fromqueue();
               strcat(hexout,hexbyte(hold,temp));
               hold[1] = fromqueue();
               strcat(hexout,hexbyte(&hold[1],temp));
               strcat(out,addoff16(hold,regs.r.ip,temp));
               break;
               
   end of comment */

/* 16 bit immediate address is valid for destination */
            case DADDR16:
               dest.where = MEMORY;  /* codegen */
               hold[0] = fromqueue();
               strcat(hexout,hexbyte(hold,temp));
               hold[1] = fromqueue();
               strcat(hexout,hexbyte(&hold[1],temp));
               if (ptrhold[0])
                  strcat(out,ptrhold);
               strcat(out,"[");
               if (prefix & SEG)
                  getseg();
               strcat(out,hnumber((hold[1] << 8) + hold[0],temp,instr_seg,regs.r.ip,startip));
               strcat(out,"]");
               dest.seg = instr_seg;
               dest.ofs = (hold[1] << 8) + hold[0];
               if (instruction.flags & BYTE)
                  dest.value = (UNSIGNED_LONG)fetchbyte(instr_seg,dest.ofs);
               else
                  dest.value = (UNSIGNED_LONG)fetchword(instr_seg,dest.ofs);
               break;

/*            case DADDR32:  back to not used again
               for (i = 0; i < 4; i++
                  {
                  hold[i] = fromqueue();
                  strcat(hexout,hexbyte(&hold[i],temp));
                  }
               strcat(out,hexlong(hold,temp));
               break;  */

            case DL:
               strcat(out,regnames[reg+DES-DAX]);
               dest.where = SEGREG;
               dest.reg = reg + 8;
               dest.value = (UNSIGNED_LONG)regs.wordarray[dest.reg];
               break;
            case DM:
               if (ptrhold[0]) /* append WORD PTR or BYTE PTR if present */
                  strcat(out,ptrhold);
               if (prefix & SEG)
                  getseg();  /* codegen */
               strcat(out,modename(hexout,hold));
               if (prefix & SEG)
                  getseg();
               if (instruction.flags & BYTE)
                  {
                  if (mod < 3)
                     {
                     dest.where = MEMORY;  /* codegen */
                     dest.seg = instr_seg;
                     dest.ofs = instr_ofs;
                     dest.value = (UNSIGNED_LONG)fetchbyte(instr_seg,instr_ofs);
                     }
                  else
                     {
                     dest.where = REGISTER;
                     dest.value = (UNSIGNED_LONG)regs.bytearray[holdreg];
                     dest.reg = holdreg;
                     }
                  }
               else
                  {
                  if (mod < 3)
                     {
                     dest.where = MEMORY;  /* codegen */
                     dest.seg = instr_seg;
                     dest.ofs = instr_ofs;
                     dest.value = (UNSIGNED_LONG)fetchword(instr_seg,instr_ofs);
                     }
                  else
                     {
                     dest.where = REGISTER;
                     dest.value = (UNSIGNED_LONG)regs.wordarray[holdreg];
                     dest.reg = holdreg;
                     }
                  }
               break;
            }
            dest.oldvalue = dest.value;
         }
      if ((instruction.source != SZ) && (instruction.dest != DZ))
         strcat(out,",");  /* if both dest and source, use a comma */
      if ((instruction.source >= SAX) && (instruction.source <= SBH))
         {  /* same as for dest above */
         strcat(out,regnames[instruction.source-SAX]);
         if (instruction.source >= SAL)
            {
            source.where = REGISTER;
            source.reg = mixbytereg[instruction.dest - SAL];
            }
         else
            if (instruction.source >= SFL)
               {
               source.where = REGISTER;
               if (instruction.source == SRAL)
                  source.reg = mixbytereg[0];
               if (instruction.source == SRAX)
                  source.reg  = 0;
               }
            else
               if (instruction.source >= SES)
                  {
                  source.where = SEGREG;
                  source.reg = instruction.source - SES + 8;
                  }
               else
                  {
                  source.where = REGISTER;
                  source.reg = instruction.source - SAX;
                  }
         if (instruction.flags & BYTE)
            source.value = (UNSIGNED_LONG)regs.bytearray[source.reg];
         else
            source.value = (UNSIGNED_LONG)regs.wordarray[source.reg];
         source.oldvalue = source.value;
         }
      else
         sourceswitch(hexout,out,ptrhold);
            /* Note:  This had to be moved to a separate procedure due to
               compiler limitations. */
      if (instruction.flags & (NOCOUNT+COUNT+INT3))
         {  /* special cases for decoding into a string */
         if (instruction.flags & NOCOUNT)
            strcat(out,",1");
         if (instruction.flags & COUNT)
            strcat(out,",CL");
         if (instruction.flags & INT3)
            strcat(out,"3");
         }
      if ((byte_one & '\375') == '\151')  /* if IMUL immediate */
         {
             /* put immediate data into dest */
         hold[0] = fromqueue();
         strcat(out,",");
         strcat(hexout,hexbyte(hold,temp));
         if (byte_one == '\151')
            {
            hold[1] = fromqueue();
            strcat(hexout,hexbyte(&hold[1],temp));
            strcat(out,numberh_word((hold[1] << 8) + hold[0],temp));
            dest.value = (UNSIGNED_LONG)((hold[1] << 8) + hold[0]);
            }
         else
            {
            if (hold[0] > 0x7f)
               {
               strcat(out,"-");
               dest.value = (-1 ^ 0xff) | hold[0];  /* that should be machine
                                                 and compiler independent */
               byte = -(hold[0] & 0x7f) + 128;
               strcat(out,numberh_byte(byte,temp));
               }
            else
               {
               strcat(out,"+");
               dest.value = (UNSIGNED_LONG)hold[0];
               strcat(out,numberh_byte(hold[0],temp));
               }
            }
         }
      }
   else  /* if not valid instruction, return it as a DB xxxx  (even
            though DB is not strictly valid for a multi-byte undefined
            operation */
      {
      strcat(out,"DB\t");
      strcat(out,hexout);
      }
   if (instruction.name == UNDEF)  /* decode status to calling routine */
      return(INVALID);
   else
      return(VALID);
   }
