/*
   This file is copyright (c) 1988 by William Couture.  All rights reserved.

   Fix PROGEND so that both end of code segment and end of data segment are
   recognized!

   4/7/92
   Un-Fix PROGEND so that entire code space is recognized!
*/

/*
   CODEGEN v1.20

   history:  too checkered to mention.

   format of .CGM files:
      char signature[4]
      unsigned int map_ofs
                   =  ofs of start of map in file
      unsigned int load_ofs
      unsigned int load_seg
      unsigned int filetype_flag
      unsigned int exe_model
      unsigned int exe_data_seg
      unsigned int flagword
                   CGM_FLAG_DOMAP     0x0001
                   CGM_FLAG_INTADDR   0x0002
      unsigned int start_offset
                   if bit 0 of flagword is set, this is significant
      unsigned int intnum_offset
                   pointer to flagword data, if bit 1 is set.
                   Structure is:
                      unsigned int num_ints
                      struct
                         {
                         unsigned char intnum
                         unsigned int intofs
                         } int_data[num_ints]
      CODEGEN map information, crunched as appropriate.

*/

#include <compiler.h>  /* Makes the source code more transportable */
#include <codegen.h>
#include <regstruc.h>
#include <opdefs.h>

#define MIN(x, y)   (((x) < (y)) ? (x) : (y))
#define MAX(x, y)   (((x) > (y)) ? (x) : (y))

extern char regnames[23][3];  /* register names (i.e. AX, BX, CH, etc) */
extern char opnames[129][7];  /* names of instructions, used for X command */
extern char *equiv_jmp_names[19];  /* alternate Jump conditional names */
extern char *dos_fn_names[0x69];  /* names of DOS functions */
extern char *video_fn_names[0x1d];  /* names of INT 10 functions */
extern char *disk_fn_names[0x18];  /* INT 13 functions */
extern char *serial_fn_names[4];  /* INT 14 functions */
extern char *cassette_fn_names[4];  /* really old functions! */
extern char *atbios_fn_names[0x12];  /* INT 15 functions */
extern char *keyboard_fn_names[0x13];  /* INT 16 functions */
extern char *parallel_fn_names[3];  /* INT 17 functions */
extern char *timer_fn_names[8];  /* INT 1A functions */
extern char *mouse_fn_names[0x1f];  /* INT 33 functions */
extern char *emm_fn_names[0x1e];  /* INT 67 functions */

extern void (*opfns[129])();  /* register tracking functions */

unsigned char alignment_byte_2;  /* use this byte for alignment */

union regunion regs;  /* See the regstruc.h file for details */

union regunion regvalues;  /* for function call determination */
union regflags regsknown;  /* regs known or unknown */

  /* regsknown.wordknown[x] : AX CX DX BX SP BP SI DI (opcode order)
     regsknown.byteknown[x] : AL AH CL CH DL DH BL BH regs.bytearray order
       These flags can be addressed with the dest.reg and source.reg values */


struct optable instruction;  /* the instruction as decoded for program
                                internal use */

struct tofrom dest,source;   /* info about the source and dest fields in the
                                current instruction, supplied by decode()
                                and decodens() */

struct symbol_table symbols[SYM_TABLE_SIZE];
unsigned int num_syms;  /* how many symbols are used */

struct cgmap_header cgm_header;

struct call_addr noret_call_addrs[NUM_NORET_CALL_ADDRS+1];
unsigned int num_call_addrs;  /* 0 = none, else index */

unsigned int num_ints_to_comment;
struct int_offset ints_to_comment[NUM_ROM_INTS];

int mod,reg,rm;  /* EA byte fields, set by decode() and decodens() */
unsigned int startip;  /* IP of the start of the instruction! */

unsigned int prefix;  /* global info on the current prefix instructions */
unsigned int hold_prefix;  /* hold info to put prefix into next instr */
unsigned int this_prefix;  /* so hold_prefix only has the last prefix */
unsigned int prefixip;  /* addr of last prefix instruction of current
                         instruction, used in case of an "interrupt" during
                         multi-prefixed instructions to emulate the 8088's
                         pushing the addr of the last prefix. See decode.c */
unsigned char byte_one;  /* the 1st byte of the decoded instruction */
unsigned char alignment_byte_1;
unsigned int memok;  /* if TRUE, have not referenced a guarded address */
unsigned int cmdptr;  /* pointer into the command line from the user */

unsigned int segoffset;  /* virtual relocation segment */
unsigned int mapspace;  /* segment used to hold the byte map of the program */
unsigned int codemap;  /* map segment corresponding to the code space */
unsigned int datamap;  /* map segment corresponding to the data space */
unsigned int memtop,memused;  /* used and available data segments */

unsigned int codesegment;  /* CODEGEN's code segment */
unsigned int datasegment;  /* CODEGEN's data segment */
unsigned int highsegment;  /* the value of the higher of the two segments */

unsigned int deadmem;  /* segment for mapping of non-existant memory */
unsigned int maxpara;  /* how much memory is available */
unsigned int memmap[16];  /* values for virtual segment mapping */

unsigned int pspaddr;  /* our PSP, used for DOS shell */
unsigned int usedmem;  /* memory used by CODEGEN */
unsigned int availmem;  /* flag: memory available for DOS shell
                                 (must be >= 64K to allow shell) */
unsigned char null_fcb[16] =
{0x00, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
 0x00, 0x00, 0x00, 0x00 };
struct
   {
   unsigned int env_seg;
   char *cmdline;
   unsigned int cmdline_seg;
   char *fcb1;
   unsigned int fcb1_seg;
   char *fcb2;
   unsigned int fcb2_seg;
   } dos4b_block = {0, "\001 \015", 0, null_fcb, 0, null_fcb, 0};
unsigned char command_name[] = "\COMMAND.COM";
unsigned char comspec_name[] = "COMSPEC=";


unsigned int progstart;   /* load address of the program */
unsigned int progend;     /* load address plus length of program */
unsigned int startloc;  /* entry point into the program */
unsigned int physical_eof;  /* The phyisical EOF is at highsegment:phyical_eof */
unsigned int initial_sp;  /* the SP given in the .EXE header */
unsigned int load_offset;  /* offset to load program at */
unsigned int memory_model;  /* 0=.COM, 1=normal .EXE, 2= cs=ds .EXE */
unsigned int file_type;  /* 0 = COM, 1 = EXE, 2 = DEVICE */
unsigned int cpu_type;  /* 0 = 8086, 1 = 80186, 2 = 80286 */
unsigned int reloc_ds;    /* data segment found via relocation */
unsigned int recom_ds;  /* recommended ds relocation */

unsigned int upperflag;   /* force output to upper case */
unsigned int hexcodecomments;  /* remove comments on code */
unsigned int hexdatacomments;  /* remove comments on data */
unsigned int doscomments;  /* assume BIOS and DOS exist */
unsigned int ascii_comments;  /* show ASCII representation of values */
unsigned int equiv_comments;  /* show DB equivalent of ASCII strings */
unsigned int show_ptrs;  /* always use BYTE PTR or WORD PTR, even if not needed */
unsigned int strings_are_labeled;  /* strings check only at labeled locations */
unsigned int use_offset;  /* use offset labels in disassembly, value is
                             either PROG_OFFSET or 0 */
unsigned int use_offset_mask;  /* mask out or keep PROG_OFFSET bit */
unsigned int label_code_offset;  /* Code labels can be offsets */
unsigned int use_tabs;  /* output spaces or tabs */
unsigned int use_term_char;  /* use special terminating char as part of string */
unsigned int term_char;
unsigned int jmp_comments;  /* give alternate Jxx names */

unsigned int do_term_tracking;  /* need to do register tracking for DOS
                                   program termination */

unsigned int multi_segment;  /* flag: more than one non-zero relocation item found */
unsigned int highseg;  /* CS or SS, whichever is higher */
unsigned int lowseg;   /* CS or SS, whichever is lower */

int fileout;                /* the output file */
unsigned int fileout_len;  /* how far along line, tabs=8 */

unsigned int myfputsok;

unsigned int pagepause;  /* pause on pages during Preview */
unsigned int linecount;  /* count of lines for pause */

unsigned int string_search;  /* true or false */
unsigned int string_found;
unsigned int search_ip;  /* starting IP for this line */
unsigned int find_str_len;

unsigned int parse_seg;  /* default segment for parse_seg_and_ofs() */

int xref_index;
int xref_file;
int do_xref;
int skip_short_xref;
unsigned int skip_short_xref_dist;
char xref_index_name[14];
char xref_file_name[14];

unsigned char instr[150];  /* decoded instruction string */
unsigned char prefix_str[150];  /* the prefix string */
unsigned char globalstr[150];  /* global temporary string */
unsigned char search_str[80];  /* search pattern */
unsigned char find_str[150];  /* buildt-up pattern */

unsigned int mixbytereg[8] =
   {
/* al  cl  dl  bl  ah  ch  dh  bh   r/m byte reg definitions 0..7 */
    0,  2,  4,  6,  1,  3,  5,  7
/* al  ah  cl  ch  dl  dh  bl  bh   regs.bytearray register order  */
   };

char tab[2] = "\t";

unsigned char nl_str[2] = "\n";  /* used frequently */
unsigned char null_str[2] = "";  /* when a null string is needed */
unsigned char onechar[2] = " ";
unsigned char shortstr[2] = " ";
unsigned char ask_option[] = "Option? ";
unsigned char ill_addr_msg[] = "Error:  illegal address";
unsigned char ill_sym_msg[] = "Illegal symbol--please try again.\n";
unsigned char num_sym_msg[] = " symbols in table.  ";
unsigned char sym_tab_full_msg[] = "Symbol table full.";
unsigned char sym_tab_empty_msg[] = "The symbol table is empty.";
unsigned char sym_addr_instr1[] = "Please enter the symbol addresses in the format\n   offset\n";
unsigned char sym_addr_instr2[] = "      or\n   segment:offset\n";
unsigned char sym_addr_instr3[] = "where segment is CS, DS, ES, SS, or a hex number, and offset is a hex number.\n";
unsigned char sym_addr_instr4[] = "The default segment is DS.\nEnter a blank line to quit.\n\n";
unsigned char enter_sym_addr_msg[] = "enter symbol address? ";
unsigned char ending_ofs_msg[] = "What is the ending offset (in hex), or blank for the starting offset? ";
unsigned char no_match_sym_msg[] = "No matching symbols.";
unsigned char init_sp_msg[] = ";\n;  INITIAL SP AT ";
unsigned char esc_to_quit_msg[] = "Press any key to try again, or Esc to quit";
unsigned char esc_to_exit_msg[] = "Press any key to try again, or Esc to exit";
unsigned char ill_resp_msg[] = "\nIllegal response.";
unsigned char start_loc_msg[] =
"For a special starting location, enter it in HEX or press Return for none.";
unsigned char start_loc_2_msg[] = "What is the starting location in HEX (0 = NONE)? ";
unsigned char end_loc_msg[] =
"For a special ending location, enter it in HEX or press Return for none.";
unsigned char end_loc_msg_2[] = "What is the ending location in HEX (Return = NONE)? ";
unsigned char which_seg_msg[] = "In which segment (Code or Data, Return = Code)? ";
unsigned char bad_start_ip_msg[] = "Warning: starting IP is outside the program.  The default IP will be used.";

unsigned char *cpu_type_names[3] = {
"8086",
"80186",
"80286"
};

unsigned char *file_type_names[3] = {
"COM/BIN",
"EXE",
"Device Driver"
};


unsigned char filename[80];  /* allow for paths */
unsigned char mapname[80];
unsigned char outfile[80];
                         /* blank for displaying a disassembled instruction */
unsigned char lowerstr[150];  /* used in myfputs() */
unsigned char diskbuf[DISKBUF_SIZE];

/*
   Currently, I don't know if the following will be used or not!
*/
unsigned int byte,flag,count,keyin;

char *hexword(), *itox();

long lseek();

char copyright[] = "CODEGEN V1.20  Copyright (c) 1988-1990 by William Couture.";

unsigned int fetchbyte(), fetchword(), peekw(), isastring();
unsigned char *strupr(), *strlwr();

main()
   {
   unsigned int temp1,temp2;
   int i;

   /* make sure this is DOS 2.0 or greater */
   cmdptr = 1;  /* use cmdptr as a temporary variable, start at 1 (pre-2.0) */
#asm
   mov ah,30h    ;function = 30h, DOS version
   int 21h       ;call DOS
   cmp al,2      ;2.0 or above?
   jnae wrongdos ;no, skip this
   mov word cmdptr_,0  ;2.0 or greater, everything ok
wrongdos:
#endasm
   if (cmdptr)  /* if pre-2.0 */
      {
      puts("Sorry, but CODEGEN requires DOS 2.0 or above.\n");
      exit();
      }
   datasegment = DATASEG;  /* get data and code seg.  Note implied small
                              memory model. */
   codesegment = CODESEG;
   highsegment = datasegment;
   if (codesegment > datasegment)  /* do not assume DS > CS */
      highsegment = codesegment;
   pspaddr = peekw(codesegment,1) + 3;
   pspaddr = peekw(codesegment,pspaddr+3);
   pspaddr = peekw(datasegment,pspaddr);  /* see the DeSmet init code to
                                             understand these 3 statements */
   mapspace = highsegment + 0x1000;  /* allow for 64K (small memory model) */
   if (mapspace & 0x0fff)
      mapspace = (mapspace & 0xf000) + 0x1000;

   do_term_tracking = TRUE;  /* default to YES */

   cpu_type =     /* 8086 default */
   term_char = 0;

   use_term_char = TRUE;
   use_tabs =
   equiv_comments =
   ascii_comments =
   hexdatacomments =
   doscomments = use_term_char;

   jmp_comments = FALSE;
   do_xref =
   label_code_offset =
   upperflag =
   strings_are_labeled =
   show_ptrs =
   hexcodecomments = jmp_comments;

   use_offset = PROG_OFFSET;  /* use this as the TRUE value */
   use_offset_mask = 0xffff;  /* keep all bits */

   do
      {
      cpu_type = 0;
      reset_params(YES);
      if ((maxpara < 0x2000) || (memused > memtop))
         {
         puts(copyright);
         puts("\n\nError:  Codegen needs a minimum of 256K of available memory beyond the\n");
         puts("        program to run!\n");
         exit(2);
         }
      } while (menu() == 'R');
   }

reset_params(map_flag)
   int map_flag;
   {
   unsigned int temp1;

   if (map_flag)
      for (temp1 = mapspace; temp1 < mapspace+0x4000; temp1 += 0x0800)
         memfill(temp1,0,0x8000,0);  /* seg, offset, len, char */
   codemap =
   datamap = mapspace;
   segoffset = mapspace + 0x2000;  /* allow for small memory model .EXE files */
   deadmem = 0xd000;  /* default to mapping "dead" memory to segment D000 */
   setmaps();  /* set map data for virtual memory and DMA mapping */
   num_syms = 0;  /* zero out the symbol table */
   num_call_addrs =
   num_ints_to_comment =
   recom_ds =
   reloc_ds =
   memory_model =
   physical_eof =
   highseg =
   lowseg =
   regs.r.cs =
   regs.r.ds =
   regs.r.es =
   regs.r.ss = num_syms;        /* make sure segs start at 0 */
   load_offset =
   startloc =
   regs.r.ip = 0x100;  /* let programs think they have a psp */
   initial_sp = 0xfffe;  /* COM initial SP */
   memok = TRUE;  /* set up default characteristics. */
   file_type = TYPE_COM;  /* default to .COM program */
   string_search =
   multi_segment = FALSE;
   mapname[0] =
   filename[0] =
   outfile[0] = '\0';
   set_processor_level(cpu_type);
   }

void menu()
   {
   unsigned int datastart, dataend, dataseg,temp;
   long ltemp;
   int i, numin, mygetsval;
   int row,col;
   char answer, ch2;  /* ch2 to keep stack aligned */
   char command_str[20];  /* holds allowable commands */

   cls();
   do
      {
      cls();
      strcpy(command_str,"WMLSCTNGPQDR");
      if (file_type != TYPE_EXE)
         command_str[10] = '\0';  /* don't need D and R */
      putsat(0,0,copyright);
      puts("\n\n  W: Workfile: ");
      puts(filename);
      puts("\n     file type: ");
      puts(file_type_names[file_type]);
      if (file_type == TYPE_EXE)
         {
         puts("\n  D: Data segment: ");
         if (memory_model == 1)
            puts("CS = CODE,    SS = DS = ES = STACK/DATA");
         else
            puts("CS = DS = ES = CODE/DATA, SS = STACK");
         puts("\n  R: stack/data Relocation: ");
         puts(itox(reloc_ds,globalstr));
         if ((recom_ds) && ((recom_ds != reloc_ds) || (reloc_ds != regs.r.ds)))
            {
            puts("    header value: ");
            puts(itox(regs.r.ds,globalstr));
            puts("  recommended value: ");
            puts(itox(recom_ds,globalstr));
            }
         }
      puts("\n  M: Map file: ");
      puts(mapname);
      puts("\n  L: Load offset: ");
      puts(itox(load_offset,globalstr));
      puts("\n  S: Starting IP: ");
      puts(itox(startloc,globalstr));
      puts("\n  C: CPU type: ");
      puts(cpu_type_names[cpu_type]);
      puts("\n  T: check for program Termination: ");
      onoroff(do_term_tracking,FALSE);
      puts("\n\n  N: handle a NO-RETURN call address");
      puts("\n  G: Goto main menu (do not map program)");
      puts("\n  P: map Program, proceede to main menu");
      if (availmem)
         {
         puts("\n  Z: execute a DOS shell");
         strcat(command_str,"Z");
         }
      puts("\n\n  Q: Quit to DOS");
      do
         {
         curpos(20,0);
         puts(ask_option);
         clear_to_bottom(20,8);
         mygets(globalstr);
         answer = toupper(globalstr[0]);
         } while (char_in_str(answer,command_str) == -1);
      for (i = 1; (globalstr[i]) && (globalstr[i] == ' '); i++)
         ;
      strcpy(instr,&globalstr[i]);  /* use instr as temp string */
      strcpy(globalstr,instr);
      switch (answer)
         {
         case 'W':  /* work file */
            cls();
            if (!globalstr[0])
               {
               putsat(2,0,"Please enter the name of the file to disassemble:");
               putsat(4,5,"? ");
               mygets(globalstr);
               }
            if (!globalstr[0])
               break;
            strupr(globalstr);
            if ((i = open(globalstr,MODE_READ)) == -1)
               {
               clear_to_bottom(5,0);
               putsat(6,0,"Error:  File not found.");
               pressany(FALSE);
               break;
               }
            numin = read(i,diskbuf,10);
            close(i);
            reset_params(YES);
            strcpy(filename,globalstr);
            if ((numin >= 2) &&
               ((*(unsigned int *)diskbuf == 0x5a4d)
                      || (*(unsigned int *)diskbuf == 0x4d5a)))
               {
               load_offset =
               startloc =
               regs.r.ip = 0;
               memory_model = 1;
               file_type = TYPE_EXE;
               }
            if ((numin == 10) && (*(unsigned int *)diskbuf == 0xffff)
                         && (*(unsigned int *)&diskbuf[2] == 0xffff))
               {
               file_type = TYPE_DEV;
               startloc = *(unsigned int *)&diskbuf[6];
               load_offset =
               regs.r.ip = 0;
               init_device_map();
               }
            if (!get_reloc_params(6))
               reset_params(NO);  /* no need to reset maps! */
            break;
         case 'M':  /* map file */
            cls();
            if (!filename[0])
               {
               putsat(2,0,"Please specify the work file before loading a map file.");
               pressany(FALSE);
               break;
               }
            if (!globalstr[0])
               {
               strcpy(instr,filename);  /* use instr as a temp string */
               for (i = 0; i < strlen(instr); i++)
                  if (instr[i] == '.')
                     instr[i] = '\0';
               strcat(instr,".CGM");
               putsat(2,0,"Press Return for ");
               puts(instr);
               puts(",\nor enter the map filename:");
               putsat(5,5,"? ");
               if (mygets(globalstr) == -1)
                  break;
               }
            if (!globalstr[0])
               strcpy(globalstr,instr);
            strupr(globalstr);
            if ((i = open(globalstr,MODE_READ)) == -1)
               {
               putsat(7,0,"Unable to open map file.");
               pressany(FALSE);
               break;
               }
            close(i);
            i = load_offset;
            if (!read_map(globalstr))
               {
               putsat(7,0,"Error:  Illegal map file.\n");
               pressany(FALSE);
               break;
               }
            strcpy(mapname,globalstr);
            if (load_offset != i)
               {
               progstart = progstart - i + load_offset;
               progend = progend - i + load_offset;
               physical_eof = physical_eof - i + load_offset;
               }
            break;
         case 'D':  /* data segments */
            memory_model = 3 - memory_model;
            break;
         case 'R':  /* relocation */
            if (!globalstr[0])
               {
               putsat(22,0,"Please enter the value for the stack/data segment relocation? ");
               mygets(globalstr);
               }
            if (globalstr[0])
               if ((i = numval(globalstr,&reloc_ds)) == FALSE)
                  {
                  putsat(23,0,"Error:  Invalid stack/data segment");
                  pressany(FALSE);
                  }
            clear_to_bottom(22,0);
            break;
         case 'L':  /* load offset */
            if (!globalstr[0])
               {
               putsat(22,0,"Please enter the load offset for this program? ");
               mygets(globalstr);
               }
            i = load_offset;
            if (globalstr[0])
               if (numval(globalstr,&load_offset) == FALSE)
                  {
                  putsat(23,0,"Error:  Invalid load offset");
                  pressany(FALSE);
                  }
               else
                  {
                  progstart = progstart - i + load_offset;
                  progend = progend - i + load_offset;
                  physical_eof = physical_eof - i + load_offset;
                  startloc =
                  regs.r.ip = load_offset;
                  }
            clear_to_bottom(22,0);
            break;
         case 'S':  /* starting IP */
            if (!globalstr[0])
               {
               putsat(22,0,"Please enter the starting IP for this program? ");
               mygets(globalstr);
               }
            if (globalstr[0])
               {
               if ((i = numval(globalstr,&startloc)) == FALSE)
                  {
                  putsat(23,0,"Error:  Invalid starting IP");
                  pressany(FALSE);
                  }
               else
                  if ((startloc < progstart) || (startloc > progend))
                     {
                     putsat(23,0,bad_start_ip_msg);
                     startloc = load_offset;
                     pressany(FALSE);
                     }
               }
            clear_to_bottom(22,0);
            break;
         case 'C':  /* CPU type */
            cpu_type++;
            if (cpu_type > 2)
               cpu_type = 0;
            set_processor_level(cpu_type);
            break;
         case 'N':
            set_noret_call_addrs();
            break;
         case 'T':
            do_term_tracking = !do_term_tracking;
            break;
         case 'G':  /* goto main menu */
         case 'P':  /* map Program */
            if (!filename[0])
               {
               putsat(22,0,"A file must be loaded before CODEGEN can proceede!");
               pressany(FALSE);
               answer = 'A';  /* something besides a 'G' or 'P' */
               }
            break;
         case 'Q': 
            cls();
            puts("\n\nQUIT: Are you sure (Return = No)? ");
            mygets(globalstr);
            if (toupper(globalstr[0]) != 'Y')
               answer = 'A';  /* something besides a 'Q' */
            break;
         case 'Z':
            cls();
            dos_shell();
            break;
         }
      } while ((answer != 'G') && (answer != 'P') && (answer != 'Q'));
   if (answer == 'Q')
      return(answer);  /* that's all, folks! */
   cls();
   if (!load_file())
      {
      putsat(9,0,"Error:  Unable to load file");
      pressany(FALSE);
      return('R');  /* return to menu! */
      }
   regs.r.ds =
   regs.r.es = regs.r.ss;
   if (memory_model == 2)
      {
      regs.r.ds =
      regs.r.es = regs.r.cs;
      }
   if (answer != 'G')  /* if not Goto main menu */
      map_it(startloc);
   strcpy(command_str,"DCFJNOSMLPWRQZ");
   if (!availmem)
      command_str[13] = '\0';  /* don't need the Z */
   do
      {
      cls();
      puts("Work file: ");
      puts(filename);
      puts("\nPlease select an option:\n\n");
      puts("  C: force a CODE area\n");
      puts("  D: force a DATA area\n");
      puts("  F: FORGET everything about an area\n");
      puts("  J: map a JUMP/CALL table\n");
      puts("  N: handle a NO-RETURN call address\n");
      puts("  O: disassembly Options menu\n");
      puts("     comment flags:  ASCII:");
      onoroff(ascii_comments,FALSE);
      puts(", hex code:");
      onoroff(hexcodecomments,FALSE);
      puts(", hex data:");
      onoroff(hexdatacomments,FALSE);
      puts("\n        equivalent DB:");
      onoroff(equiv_comments,FALSE);
      puts(", interrupt:");
      onoroff(doscomments,FALSE);
      puts("\n     other flags:  upper case:");
      onoroff(upperflag,FALSE);
      puts(", use TABs:");
      onoroff(use_tabs,FALSE);
      puts(", strings must be labled:");
      onoroff(strings_are_labeled,FALSE);
      puts("\n        use PTR overrides:");
      onoroff(show_ptrs,FALSE);
      puts(", use OFFSETs:");
      onoroff(use_offset,FALSE);
      puts(", code labels as OFFSETs:");
      onoroff(label_code_offset,FALSE);
      puts("\n        alternate JMP names:");
      onoroff(jmp_comments,FALSE);
      puts(", string termination char: ");
      if (use_term_char)
         puts(hexbyte(&term_char,globalstr));
      else
         puts("none");
      puts("\n  S: SYMBOL table options  ");
      if (num_syms)
         {
         itoa(num_syms,globalstr);
         puts(globalstr);
         puts(num_sym_msg);
         }
      puts("\n  M: save the current MAP (can be re-loaded later for furthur work)\n");
      puts("  L: LOCATE a string in the disassembled file (Esc to quit)\n");
      puts("  P: PREVIEW the disassembled code to the screen (Esc to quit)\n");
      puts("  W: WRITE the disassembled code to a file (Esc to quit)\n\n");
      puts("  R: RESTART Codegen\n");
      puts("  Q: QUIT back to dos\n");
      if (availmem)
         puts("  Z: execute a DOS shell");
      do
         {
         curpos(24,0);
         puts(ask_option);
         clear_to_bottom(24,8);
         mygets(globalstr);
         answer = toupper(globalstr[0]);
         } while (char_in_str(answer,command_str) == -1);
      switch (answer)
         {
         case 'D':
         case 'F':
            cls();
            if (answer == 'D')
               puts("Force a DATA area:");
            else
               puts("FORGET everything about an area:");
            puts("\n\nThe default segment is DS.\n");
            puts("Enter a blank for the starting address to quit.");
            do
               {
               do
                  {
                  putsat(5,0,"What is the starting address (in hex, use SEG:OFS if needed)? ");
                  mygets(globalstr);
                  parse_seg = regs.r.ds;  /* default parsing segment */
                  if ((globalstr[0])
                         && ((temp = parse_seg_and_ofs(&dataseg,&datastart,&i,globalstr)) == FALSE))
                     {
                     clear_to_bottom(5,62);
                     putsat(7,0,ill_addr_msg);
                     }
                  } while ((globalstr[0]) && (temp == FALSE));
               if (globalstr[0])
                  {
                  do
                     {
                     putsat(7,0,ending_ofs_msg);
                     mygets(globalstr);
                     if (!globalstr[0])
                        hexword(&datastart,globalstr);
                     if ((temp = numval(globalstr,&dataend)) == FALSE)
                        {
                        clear_to_bottom(7,70);
                        putsat(9,0,ill_addr_msg);
                        }
                     } while (!temp);
                  clear_to_bottom(9,0);
                  if (datastart <= dataend)
                     {
                     temp = (dataseg == regs.r.cs) ? codemap : datamap;
                     if (answer == 'F')
                        {
                        if (dataend == 0xffff)
                           dataend--;
                        for (count = datastart; count <= dataend; count++)
                           _poke(PROG_UNKNOWN,count,temp);
                        }
                     else
                        {
                        for (count = datastart; count <= dataend; count++)
                           _poke(_peek(count,temp) & (~(PROG_CODE|PROG_LABEL|PROG_KNOWN|PROG_OFFSET)),count,temp);
                        puts("Do you want to force a label here (Return = No)? ");
                        mygets(globalstr);
                        if (toupper(globalstr[0]) == 'Y')
                           if (dataseg == regs.r.cs)
                              SETCFLAG(datastart,PROG_DATA);
                           else
                              SETDFLAG(datastart,PROG_DATA);
                        puts("Do you want to force this to be Word data (Return = No)? ");
                        mygets(globalstr);
                        if (toupper(globalstr[0]) == 'Y')
                           if (dataseg == regs.r.cs)
                              SETCFLAG(datastart,PROG_WORD);
                           else
                              SETDFLAG(datastart,PROG_WORD);
                        }
                     globalstr[0] = 'a';  /* something besides a blank! */
                     clear_to_bottom(5,0);
                     }
                  else
                     {
                     puts("The ending address must be equal to or larger than the starting address.\n");
                     pressany(FALSE);
                     clear_to_bottom(5,0);
                     }
                  }
               } while (globalstr[0]);
            break;
         case 'C':
            do
               {
               cls();
               puts("Force a CODE area:\n\nEnter a blank for the starting address to quit.");
               do
                  {
                  putsat(4,0,"What is the starting address (in hex)? ");
                  mygets(globalstr);
                  if ((globalstr[0]) && ((temp = numval(globalstr,&datastart)) == FALSE))
                     {
                     clear_to_bottom(4,39);
                     putsat(6,0,ill_addr_msg);
                     }
                  } while ((globalstr[0]) && (!temp));
               if (globalstr[0])
                  if ((datastart >= progstart) && (datastart < progend))
                     map_it(datastart);
               } while (globalstr[0]);
            break;
         case 'J':
            map_jump_table();
            break;
         case 'N':
            set_noret_call_addrs();
            break;
         case 'M':
            write_map();
            break;
         case 'O':
            options_menu();
            break;
         case 'S':
            symbol_menu();
            break;
         case 'W':
            write_file();
            break;
         case 'P':
            preview_file();
            break;
         case 'L':
            locate_string();
            break;
         case 'Q':
            cls();
            puts("\n\nQUIT: Are you sure (Return = No)? ");
            mygets(globalstr);
            if (toupper(globalstr[0]) != 'Y')
               answer = 'A';  /* something besides a 'Q' */
            break;
         case 'R':
            cls();
            puts("\n\nRESTART: Are you sure (Return = No)? ");
            mygets(globalstr);
            if (toupper(globalstr[0]) != 'Y')
               answer = 'A';  /* something besides a 'Q' */
            break;
         case 'Z':
            cls();
            dos_shell();
            break;
         }
      } while ((answer != 'Q') && (answer != 'R'));
   return(answer);
   }

options_menu()
   {
   char answer, ch2;  /* ch2 to keep stack aligned */

   do
      {
      cls();
      puts("Disassembly Options:\n\nDo you want to:\n\n");
      puts("  A: ASCII comments (currently");
      onoroff(ascii_comments,TRUE);
      puts("  C: hex comments on disassembled CODE (currently");
      onoroff(hexcodecomments,TRUE);
      puts("  D: hex comments on disassembled DATA (currently");
      onoroff(hexdatacomments,TRUE);
      puts("  E: EQUIVALENT DB comments to strings (currently");
      onoroff(equiv_comments,TRUE);
      puts("  I: DOS and BIOS INTERRRUPT comments (currently");
      onoroff(doscomments,TRUE);
      puts("  J: give alternate conditional JMP names (currently");
      onoroff(jmp_comments,TRUE);
      puts("  L: code LABELS can be OFFSETS (currently");
      onoroff(label_code_offset,TRUE);
      puts("  O: OFFSET usage (currently");
      onoroff(use_offset,TRUE);
      puts("  P: always generate BYTE PTR and WORD PTR, even if not\n");
      puts("     strictly necessary (currently");
      onoroff(show_ptrs,TRUE);
      puts("  S: Strings must start at a label (currently");
      onoroff(strings_are_labeled,TRUE);
      puts("  T: use TABs in output (currently");
      onoroff(use_tabs,TRUE);
      puts("  U: UPPER case output of the disassembled file (currently");
      onoroff(upperflag,TRUE);
      puts("  Z: string terminator char (currently ");
      if (use_term_char)
         puts(hexbyte(&term_char,globalstr));
      else
         puts("none");
      puts(")\n\n  R: Return to the main menu (Default if Return is pressed)\n");
      do
         {
         curpos(22,0);
         puts(ask_option);
         clear_to_bottom(22,8);
         mygets(globalstr);
         answer = toupper(globalstr[0]);
         if (answer == '\0')
            answer = 'R';
         } while (char_in_str(answer,"ACDEIJLOPSTUZR") == -1);
      switch (answer)
         {
         case 'U':
            upperflag = !upperflag;
            break;
         case 'C':
            hexcodecomments = !hexcodecomments;
            break;
         case 'D':
            hexdatacomments = !hexdatacomments;
            break;
         case 'I':
            doscomments = !doscomments;
            break;
         case 'A':
            ascii_comments = !ascii_comments;
            break;
         case 'E':
            equiv_comments = !equiv_comments;
            break;
         case 'O':
            use_offset = use_offset ? 0 : PROG_OFFSET;
            use_offset_mask = use_offset ? 0xffff : ~PROG_OFFSET;
            break;
         case 'L':
            label_code_offset = !label_code_offset;
            break;
         case 'P':
            show_ptrs = !show_ptrs;
            break;
         case 'S':
            strings_are_labeled = !strings_are_labeled;
            break;
         case 'T':
            use_tabs = !use_tabs;
            break;
         case 'J':
            jmp_comments = !jmp_comments;
            break;
         case 'Z':
            use_term_char = !use_term_char;
            if (use_term_char)
               {
               putsat(23,0,"What is the terminating char, in hex (Return = 0)? ");
               mygets(globalstr);
               if (numval(globalstr,&term_char) == FALSE)
                  term_char = 0;
               clear_to_bottom(23,0);
               }
            break;
         }
      } while (answer != 'R');
   }

void create_disassembled_file(use_screen,preview_start,end_loc,use_data_seg)
   int use_screen, preview_start, end_loc, use_data_seg;
   {
   char hexstr[22],prefix_hex[22],asciistr[8],jmpstr[10];
   int byte,byteorword,decodeok,len1;
   unsigned int count,temp,temp2,last_temp;
   int onceflag;

   REGS_UNKNOWN;
   fileout_len = 0;  /* start off clean */
   myfputsok = TRUE;  /* anything but -1 */
   search_ip = 0;  /* for Locate (search for) string */
   if (use_screen)
      {
      if (!string_search)
         {
         if ((fileout = fopen("CON","w")) == 0)
            {
            puts("\n\nThe screen cannot be accessed for output.\n");
            pressany(FALSE);
            return;
            }
         }
      }
   else
      {
      puts("\nCreating output file: ");
      puts(outfile);
      if ((fileout = fopen(outfile,"w")) == 0)
         {
         puts("\n\nThe output file ");
         puts(outfile);
         puts(" cannot be created.\n");
         pressany(FALSE);
         return;
         }
      }
   xref_index =
   xref_file = ERR;  /* flag they do not exist */
   if (do_xref)
      {
      strcpy(xref_index_name,"XREFINDX.00");
      strcpy(xref_file_name,"XREFFILE.00");
      if ((xref_index = create_unique(xref_index_name)) != ERR)
         {
         if ((xref_file = create_unique(xref_file_name)) == ERR)
            {
            close(xref_index);
            unlink(xref_index_name);
            xref_index = ERR;
            }
         }
      if (xref_index == ERR)
         {
         do_xref = FALSE;  /* can't do xref! */
         puts("Unable to create cross-reference files.  Cross-refernece not generated.");
         pressany(FALSE);
         }
      }
   if (!preview_start)
      {
      onechar[0] = ':';
      myfputs("\tPAGE\t,132");
      myfputs("\n;\n;  NAME: ");
      myfputs(filename);
      myfputs("\n;\n;  PROGRAM TYPE: ");
      myfputs(file_type_names[file_type]);
      myfputs("\n;\n;  CPU TYPE: ");
      myfputs(cpu_type_names[cpu_type]);
      myfputs("\n;\n;  PROGRAM LOADED AT ");
      temp = 0;
      myfputs(hexword(&temp,globalstr));
      myfputs(onechar);
      myfputs(hexword(&load_offset,globalstr));
      myfputs("\n;\n;  PHYSICAL EOF AT ");
      myfputs(hexword(&highseg,globalstr));
      myfputs(onechar);
      myfputs(hexword(&physical_eof,globalstr));
      myfputs("\n;\n;  PROGRAM ENTRY POINT AT ");
      myfputs(hexword(&regs.r.cs,globalstr));
      myfputs(onechar);
      myfputs(hexword(&startloc,globalstr));
      myfputs("\n;\n");
      if (file_type == TYPE_EXE)
         {
         myfputs(";  INITIAL STACK AT ");
         myfputs(hexword(&regs.r.ss,globalstr));
         myfputs(onechar);
         myfputs(hexword(&initial_sp,globalstr));
         myfputs("\n;\n");
         myfputs("DATAFUN\tSEGMENT\t");
         if (regs.r.ds == regs.r.cs)
            myfputs("STACK");
         else
            myfputs("BYTE\tPUBLIC");
         myfputs("\nDATAFUN\tENDS\nCODEFUN\tSEGMENT\tBYTE\tPUBLIC\nCODEFUN\tENDS\n");
         }
      }
   if ((file_type == TYPE_EXE) && (codemap > datamap)
                 && ((!preview_start) || (use_data_seg)))
      {
      myfputs("DATAFUN\tSEGMENT\n");
      temp = (end_loc == progend) ? (regs.r.cs - regs.r.ss) * 16 : end_loc;
      for (count = preview_start; count < temp; count++)
         {
         if ((create_data_output(datamap,regs.r.ss,&count,temp,use_screen)) ||
                                          (gchar() == '\033'))
            {
            if (!string_search)
               cleanup_exit_disassemble();
            return;
            }
         }
      myfputs("DATAFUN\tENDS\n");
      }
   if (!preview_start)
      {
      if (file_type == TYPE_EXE)
         {
         if ((file_type == TYPE_EXE) && (regs.r.cs) && (regs.r.ss))
            {
            count = 0;
            temp = (codemap > datamap) ? codemap : datamap;
            temp += (progend + 0x0f) >> 4;
            temp2 = (regs.r.cs > regs.r.ss) ? regs.r.ss : regs.r.cs;
            temp2 <<= 4;
            myfputs("MISSING\tSEGMENT\n");
            if (create_data_output(temp,0,&count,temp2,use_screen))
               {
               if (!string_search)
                  cleanup_exit_disassemble();
               return;
               }
            myfputs("MISSING\tENDS\n");
            }
         myfputs("CODEFUN\tSEGMENT\n");
/*         if (highseg == lowseg)
            myfputs("DATAFUN\tSEGMENT\n"); */
         myfputs("ASSUME\tCS:CODEFUN,DS:");
         if (regs.r.ds == regs.r.cs)
            myfputs("CODEFUN,ES:CODEFUN,SS:");
         else
            myfputs("DATAFUN,ES:DATAFUN,SS:");
         if (regs.r.ss == regs.r.cs)
            myfputs("CODEFUN\n");
         else
            myfputs("DATAFUN\n");
         }
      else
         myfputs("FUN\tSEGMENT\nASSUME\tCS:FUN,DS:FUN,ES:FUN,SS:FUN\n");
      onceflag = TRUE;
      for (count = 0; (count < progstart) && (myfputsok != -1); count++)
         if ((temp = PEEK(codemap,count)) & (PROG_DATA|PROG_LABEL))
            {
            search_ip = count;
            if (onceflag)
               {
               if (file_type == TYPE_EXE)
                  myfputs(";\n;  CODE SEGMENT REFERENCES BEFORE THE START OF CODE SPACE\n;\n");
               else
                  myfputs(";\n;  REFERENCES BEFORE THE START OF CODE SPACE\n;\n");
               onceflag = FALSE;
               }
            myfputs("\tORG\t");
            myfputs(numberh_word(count,globalstr));
            myfputs(nl_str);
            myfputs(hnumber(count,globalstr,regs.r.cs,0xffff,0xffff));
            myfputs("\tLABEL\t");
            if (temp & PROG_LABEL)
               myfputs("NEAR\n");
            else
               if (temp & PROG_WORD)
                  myfputs("WORD\n");
               else
                  myfputs("BYTE\n");
            if (gchar() == '\033')
               {
               if (!string_search)
                  cleanup_exit_disassemble();
               return;
               }
            }
      onceflag = TRUE;
      for (count = progstart; (count < end_loc) && (myfputsok != -1); count++)
         if ((PEEK(codemap,count) & (PROG_CODE | PROG_DATA)) == (PROG_CODE | PROG_DATA))
                                         /* if code AND data */
            {
            search_ip = count;
            if (onceflag)
               {
               myfputs(";\n;  DATA REFERENCES TO CODE SPACE ADDRESSES\n;\n");
               onceflag = FALSE;
               }
            myfputs(";\tORG\t");
            myfputs(numberh_word(count,globalstr));
            myfputs(nl_str);
            myfputs(";");
            myfputs(hnumber(count,globalstr,regs.r.cs,0xffff,0xffff));
            myfputs("\tLABEL\t");
            if (PEEK(codemap,count) & PROG_WORD)
               myfputs("WORD\n");
            else
               myfputs("BYTE\n");
            if (gchar() == '\033')
               {
               if (!string_search)
                  cleanup_exit_disassemble();
               return;
               }
            }
      myfputs(";\n;  START OF PROGRAM\n;\n\tORG\t");
      myfputs(numberh_word(progstart,globalstr));
      myfputs(nl_str);
      }
   count = (progstart > preview_start) ? progstart : preview_start;
   if ((preview_start) && (use_data_seg))
      count = end_loc;  /* don't go into code segment */
   for ( ; (count < end_loc) && (myfputsok != -1); )
      {
      byte = PEEK(codemap,count);
      if (byte & PROG_CODE)
         {
         if (num_ints_to_comment)
            {
            onceflag = TRUE;
            temp2 = FALSE;
            for (temp = 0; temp < num_ints_to_comment; temp++)
               if (ints_to_comment[temp].intofs == count)
                  {
                  if (onceflag)
                     {
                     myfputs(";\n");
                     onceflag = FALSE;
                     }
                  myfputs(";  INTERRUPT ");
                  myfputs(hexbyte(&ints_to_comment[temp].intnum,globalstr));
                  myfputs(nl_str);
                  temp2 = TRUE;  /* found an interrupt to comment */
                  }
            if (temp2)
               myfputs(";\n");
            }
         search_ip =
         regs.r.ip = count;
/*         if (byte & PROG_LABEL) */
         if ((byte & PROG_LABEL) || (byte & PROG_DATA))
            {
            myfputs(hnumber(count,globalstr,regs.r.cs,0xffff,0xffff));
            myfputs(":\n");
            if (byte & PROG_LABEL)
               REGS_UNKNOWN;
            if (!use_screen)
               puts(".");
            }
         prefix_hex[0] =
         prefix_str[0] = '\0';  /* no prefix string */
         this_prefix =
         hold_prefix =
         prefix = 0;  /* clear out prefixes */
         dest.where =
         source.where = MEM_REG_UNKNOWN; /* MEMORY; */
                                /* default to UNKNOWN source/destination */
         search_ip =
         last_temp =
         startip = regs.r.ip;  /* local for location of all prefixes, no int */
         do
            {
            temp = regs.r.ip;
            decodeok = decode(hexstr,instr);
            if (instruction.flags & PREFIX)
               {
               if (hold_prefix)
                  {
                  put_modified_instr(prefix_str,null_str,null_str,startip);
                  comment_instr(prefix_hex,null_str,null_str,last_temp);
                  myfputs(nl_str);
                  }
               strcpy(prefix_hex,hexstr);
               strcpy(prefix_str,instr);  /* save this instruction */
               hold_prefix = this_prefix;
               last_temp = temp;   /* save this offset */
               }
            else
               {
               if (hold_prefix & REP)
                  {
                  if (char_in_str(byte_one,"\154\155\156\157\244\245\246\247\252\253\254\255\256\257") == -1)
                        /* if a repeat and not a string instruction */
                     {
                     put_modified_instr(prefix_str,null_str,null_str,startip);
                     comment_instr(prefix_hex,null_str,null_str,last_temp);
                     myfputs(nl_str);
                     hold_prefix = 0;
                     }
                  else
                     {
                     strcat(prefix_str,&instr[1]);
                     strcpy(instr,prefix_str);
                     strcat(prefix_hex,hexstr);
                     strcpy(hexstr,prefix_hex);
                     temp = last_temp;  /* show location of prefix! */
                     }
                  }
               if (hold_prefix & SEG)
                  {
                  if ((temp2 = copyto(globalstr,instr,'[')) != -1)
                     {
                     prefix_str[4] = '\0';    /* eliminate trailing tab */
                     strcpy(&globalstr[temp2-1],&prefix_str[1]); /* and starting tab */
                     strcat(globalstr,&instr[temp2-1]);  /* add rest of str */
                     strcpy(instr,globalstr);  /* and back to instr */
                     strcat(prefix_hex,hexstr);
                     strcpy(hexstr,prefix_hex);
                     temp = last_temp;  /* show location of prefix */
                     }
                  else
                     {
                     put_modified_instr(prefix_str,null_str,null_str,startip);
                     comment_instr(prefix_hex,null_str,null_str,last_temp);
                     myfputs(nl_str);
                     }
                  }
               if (hold_prefix & LOCK_PREFIX)
                  {
                  strcat(prefix_str,&instr[1]);
                  strcpy(instr,prefix_str);
                  strcat(prefix_hex,hexstr);
                  strcpy(hexstr,prefix_hex);
                  temp = last_temp;  /* show location of prefix! */
                  }
               put_modified_instr(instr,asciistr,jmpstr,startip);
                                               /* use ip to change count */
               if ((doscomments) && (instruction.name == INT))
                  puts_dos_comments(temp,hexstr);
               else
                  comment_instr(hexstr,asciistr,jmpstr,temp);
               if (decodeok)
                  (*opfns[instruction.name])();  /* go do register tracking */
               if ((byte_one == 0xeb) && (fetchbyte(regs.r.cs,regs.r.ip) == 0x90)
                      && (PEEK(codemap,regs.r.ip) == PROG_UNKNOWN))
                  {
                  if (hexcodecomments)
                     {
                     temp = regs.r.ip;
                     decode(hexstr,instr);
                     myfputs(nl_str);
                     put_modified_instr(instr,null_str,null_str,startip);
                     comment_instr(hexstr,null_str,null_str,temp);
                     }
                  else
                     {
                     regs.r.ip++;
                     myfputs("\n\tDB\t90H\t\t\t\t;NOP PROBABLY DUE TO MASM");
                     }
                  }
               myfputs(nl_str);
               }
            } while ((instruction.flags & PREFIX) && (decodeok));
         count = regs.r.ip;
         }
      else
         {
         if (create_data_output(codemap,regs.r.cs,&count,end_loc,use_screen))
            {
            if (!string_search)
               cleanup_exit_disassemble();
            return;
            }
         }
      if (gchar() == '\033')
         {
         if (!string_search)
            cleanup_exit_disassemble();
         return;
         }
      }
   onceflag = TRUE;
   if (!preview_start)
      {
      for (count = end_loc; (count < 0xffff) && (myfputsok != -1); count++)
         if ((temp = PEEK(codemap,count)) & (PROG_DATA | PROG_LABEL))
            {
            search_ip = count;
            if (onceflag)
               {
               if (file_type == TYPE_EXE)
                  myfputs(";\n;  CODE SEGMENT REFERENCES AFTER THE END OF CODE SPACE\n;\n");
               else
                  myfputs(";\n;  REFERENCES AFTER THE END OF CODE SPACE\n;\n");
               onceflag = FALSE;
               }
            myfputs("\tORG\t");
            myfputs(numberh_word(count,globalstr));
            myfputs(nl_str);
            myfputs(hnumber(count,globalstr,regs.r.cs,0xffff,0xffff));
            myfputs("\tLABEL\t");
            if (temp & PROG_LABEL)
               myfputs("NEAR\n");
            else
               if (temp & PROG_WORD)
                  myfputs("WORD\n");
               else
                  myfputs("BYTE\n");
            if (gchar() == '\033')
               {
               if (!string_search)
                  cleanup_exit_disassemble();
               return;
               }
            }
      if (file_type == TYPE_EXE)
         myfputs("CODEFUN\tENDS\n");
      else
         myfputs("FUN\tENDS\n");
      }
   if ((file_type == TYPE_EXE) && (codemap <= datamap) && (highseg != lowseg)
              && ((!preview_start) || (use_data_seg)))
      {
      myfputs("DATAFUN\tSEGMENT\n");
      temp2 = (end_loc == progend) ? 0xfffe : end_loc;
      for (count = preview_start; (count < temp2) && (myfputsok != -1); count++)
         {
         if ((create_data_output(datamap,regs.r.ss,&count,temp2,use_screen)) ||
                   (gchar() == '\033'))
            {
            if (!string_search)
               cleanup_exit_disassemble();
            return;
            }
         }
      myfputs("DATAFUN\tENDS\n");
      }
   if (!preview_start)
      {
      myfputs("\tEND\t");
      myfputs(hnumber(startloc,globalstr,regs.r.cs,0xffff,0xffff));
      myfputs(nl_str);
      }
   cleanup_exit_disassemble();
   }

cleanup_exit_disassemble()
   {
   if ((myfputsok == -1) && (!pagepause))
      {
      puts("\n\nError on ouput:  disk error or disk full.\n\nFile creation aborted.\n\n");
      pressany(FALSE);
      }
   if ((xref_index != ERR) && (!do_xref))
      {
      puts("\n\nError in cross reference:  disk error or disk full.\n\nCross reference aborted.\n\n");
      pressany(FALSE);
      }
   if (do_xref)
      show_xref();  /* turns do_xref off */
   if (xref_index != ERR)
      {
      close(xref_index);
      close(xref_file);
      unlink(xref_index_name);
      unlink(xref_file_name);
      }
   if (!string_search)
      fclose(fileout);
   }

int create_data_output(whichmap,whichseg,cnt,maxcount,use_screen)
   unsigned int whichmap, whichseg;
   unsigned int *cnt;
   unsigned int maxcount,use_screen;
   {
   int byte,isitword,is_address,decodeok,num_bytes;
   unsigned int count,count2,pos,temp,len,temp2;
   char data_bytes[20], *dbytes;

   count = *cnt;
   data_bytes[0] = ' ';
   is_address =
   isitword = FALSE;
   byte = PEEK(whichmap,count) & use_offset_mask;
   do
      {
      if (gchar() == '\033')
         return(TRUE);  /* Return because ESC was pressed */
      if (byte)
         isitword = byte & PROG_WORD;  /* if unknown, leave last defn alone */
      is_address = byte & PROG_ADDRESS;  /* address does not propagate */
      search_ip =
      pos = count;
      if (byte & (PROG_DATA|use_offset))
         {
         myfputs(hnumber(count,globalstr,whichseg,0xffff,0xffff));
         if (strlen(globalstr) > 7)
            {
            if ((file_type != TYPE_EXE) || (whichseg == regs.r.cs))
               myfputs(nl_str);
            else
               {
               myfputs("\tLABEL\t");
               if (isitword)
                  myfputs("WORD\n");
               else
                  myfputs("BYTE\n");
               }
            }
         if (!use_screen)
            puts(".");
         }
      if ((count2=isastring(count,whichmap,whichseg,maxcount))
               && ((!strings_are_labeled) ||
                               ((strings_are_labeled)
                                    && (byte & (PROG_DATA|use_offset)))))
         {
         myfputs("\tDB\t");
         search_ip =
         temp = count;
         while (count < count2)
            {
            if (isaprint(onechar[0]=fetchbyte(whichseg,count++)))
               {
               if (temp+1 == count)
                  myfputs("\"");
               if (myfputsok != -1)
                  myfputs2(onechar,fileout);
               if (((count-temp) % 32 == 0) || (count == count2))
                  {
                  put_data_comment(temp,count,count2,pos,whichseg);
                  search_ip =
                  pos =
                  temp = count;
                  if (gchar() == '\033')
                     return(TRUE);  /* Return because ESC was pressed */
                  }
               }
            else
               {
               if (temp+1 < count)
                  put_data_comment(temp,count-1,count2,pos,whichseg);
               len = 4;
               search_ip =
               pos =
               temp = --count;  /* backup to where this byte came from */
               do
                  {
                  pos++;
                  if (temp != count)
                     {
                     myfputs(",");
                     len++;
                     }
                  count++;
                  myfputs(numberh_byte(onechar[0],globalstr));
                  len += strlen(globalstr);
                  } while ((!isaprint(onechar[0]=fetchbyte(whichseg,count)))
                               && ((count-temp) % 8)  && (count < count2));
               if (hexdatacomments)
                  {
                  tabout(56);
                  myfputs(hexword(&temp,globalstr));
                  }
               if (gchar() == '\033')
                  return(TRUE);  /* Return because ESC was pressed */
               if (count < count2)
                  myfputs("\n\tDB\t");
               search_ip =
               temp = count;
               }
            }
          myfputs(nl_str);
         }
      else
         {
         if (!checkdup(isitword,is_address,&count,pos,maxcount,whichmap,whichseg))
            {
            if ((isitword) 
                     && ((PEEK(whichmap,count+1)&use_offset_mask) == PROG_UNKNOWN)
                     && (count < maxcount-1))
               {
               myfputs("\tDW\t");
               dbytes = &data_bytes[1];
               temp2 = fetchword(whichseg,count);
               *(dbytes++) = showbyte(temp2 & 0xff);
               *(dbytes++) = showbyte((temp2 >> 8) & 0xff);
               if (is_address)
                  myfputs(hnumber(temp2,globalstr,regs.r.cs,whichseg,count));
               else
                  myfputs(numberh_word(temp2,globalstr));
               count += 2;
               len = 4 + strlen(globalstr);
               while (((PEEK(whichmap,count)&use_offset_mask) == PROG_UNKNOWN)
                           && ((PEEK(whichmap,count+1)&use_offset_mask) == PROG_UNKNOWN)
                           && (count < maxcount-1)
                           && ((count-pos) % 8))
                  {
                  myfputs(",");
                  temp2 = fetchword(whichseg,count);
                  *(dbytes++) = showbyte(temp2 & 0xff);
                  *(dbytes++) = showbyte((temp2 >> 8) & 0xff);
                  myfputs(numberh_word(temp2,globalstr));
                  len += 1 + strlen(globalstr);
                  count += 2;
                  }
               if (hexdatacomments)
                  {
                  tabout(56);
                  myfputs(hexword(&pos,globalstr));
                  *dbytes = '\0';
                  if (myfputsok != -1)
                     myfputs2(data_bytes,fileout);
                  }
               }
            else
               {
               myfputs("\tDB\t");
               dbytes = &data_bytes[1];
               temp2 = fetchbyte(whichseg,count++);
               *(dbytes++) = showbyte(temp2);
               myfputs(numberh_byte(temp2,globalstr));
               len = 4 + strlen(globalstr);
               while (((PEEK(whichmap,count)&use_offset_mask) == PROG_UNKNOWN)
                            && (count < maxcount)
                            && ((count-pos) % 8)
                            && ((strings_are_labeled) || (!isastring(count,whichmap,whichseg,maxcount))))
                  {
                  myfputs(",");
                  temp2 = fetchbyte(whichseg,count++);
                  *(dbytes++) = showbyte(temp2);
                  myfputs(numberh_byte(temp2,globalstr));
                  len += 1 + strlen(globalstr);
                  }
               if (hexdatacomments)
                  {
                  tabout(56);
                  myfputs(hexword(&pos,globalstr));
                  *dbytes = '\0';
                  if (myfputsok != -1)
                     myfputs2(data_bytes,fileout);
                  }
               }
            }
         myfputs(nl_str);
         }
      byte = PEEK(whichmap,count) & use_offset_mask;
      } while ((!(byte & PROG_CODE)) && (count < maxcount) && (myfputsok != -1));
   *cnt = count;
   return(FALSE);  /* no ESC was pressed */
   }

void put_data_comment(temp,count,count2,pos,whichseg)
   unsigned int temp, count, count2, pos, whichseg;
   {
   int temp2,len;

   temp2 = temp;
   len = 5;
   myfputs("\"");     /* need to end the string regardless */
   if (equiv_comments)
      {
      myfputs("\n;\tDB\t");
      while (temp < count)
         {
         myfputs(numberh_byte(fetchbyte(whichseg,temp++),globalstr));
         len += strlen(globalstr);
         if (count != temp)
            if ((temp-temp2) % 8 == 0)
               {
               if (hexdatacomments)
                  {
                  tabout(56);
                  myfputs(hexword(&pos,globalstr));
                  }
               len = 5;
               myfputs("\n;\tDB\t");
               pos += 8;
               }
            else
               {
               myfputs(",");
               len++;
               }
         }
      if ((len > 5) && (hexdatacomments))
         {
         tabout(56);
         myfputs(hexword(&pos,globalstr));
         }
      }
   if (count < count2)
      myfputs("\n\tDB\t");
   }

int checkdup(isitword,is_address,cnt,pos,maxcount,whichmap,whichseg)
   int isitword,is_address;
   unsigned int *cnt;
   unsigned int pos, maxcount, whichmap, whichseg;
   {
   unsigned int count,dupcount,value,len;
   unsigned char comment[4];

   count = *cnt;
   dupcount = 1;
   if (isitword)
      {
      if ((PEEK(whichmap,count+1)&use_offset_mask) == PROG_UNKNOWN)
         {
         value = fetchword(whichseg,count);
         count += 2;
         while (((PEEK(whichmap,count)&use_offset_mask) == PROG_UNKNOWN)
                     && ((PEEK(whichmap,count+1)&use_offset_mask) == PROG_UNKNOWN)
                     && (count < maxcount-1)
                     && (fetchword(whichseg,count) == value))
            {
            dupcount++;
            count += 2;
            }
         }
      if (dupcount > 2)
         {
         myfputs("\tDW\t");
         myfputs(numberh_word(dupcount,globalstr));
         len = 11 + strlen(globalstr);
         myfputs(" DUP (");
         if (is_address)
            myfputs(hnumber(value,globalstr,regs.r.cs,whichseg,*cnt));
         else
            myfputs(numberh_word(value,globalstr));
         len += strlen(globalstr);
         myfputs(")");
         if (hexdatacomments)
            {
            tabout(56);
            myfputs(hexword(&pos,globalstr));
            comment[0] = ' ';
            comment[1] = showbyte(value & 0xff);
            comment[2] = showbyte((value >> 8) & 0xff);
            comment[3] = '\0';
            if (myfputsok != -1)
               myfputs2(comment,fileout);
            }
         *cnt += dupcount + dupcount;
         return(TRUE);
         }
      return(FALSE);
      }
   else
      {
      value = fetchbyte(whichseg,count++);
      while (((PEEK(whichmap,count)&use_offset_mask) == PROG_UNKNOWN)
                  && (count < maxcount)
                  && (fetchbyte(whichseg,count++) == value))
            dupcount++;
      if (dupcount > 4)
         {
         myfputs("\tDB\t");
         myfputs(numberh_word(dupcount,globalstr));
         len = 11 + strlen(globalstr);
         myfputs(" DUP (");
         myfputs(numberh_byte(value,globalstr));
         len += strlen(globalstr);
         myfputs(")");
         if (hexdatacomments)
            {
            tabout(56);
            myfputs(hexword(&pos,globalstr));
            comment[0] = ' ';
            comment[1] = showbyte(value);
            comment[2] = '\0';
            if (myfputsok != -1)
               myfputs2(comment,fileout);
            }
         *cnt += dupcount;
         return(TRUE);
         }
      return(FALSE);
      }
   }

int put_modified_instr(instr,asciistr,jmpstr,loc)  /* modify IP to change count */
   char *instr, *asciistr, *jmpstr;
   unsigned int loc;
   {
   unsigned int index,count,value,tempmap,temp;
   int i;

   jmpstr[0] =
   asciistr[0] = '\0';
   index = copyto(globalstr,instr,'\t');
   index += copyto(&globalstr[index],&instr[index],'\t');
   myfputs(globalstr);
   switch (instruction.name)
      {
      case JA:
      case JAE:
      case JCXZ:
      case JGE:
      case JG:
      case JNA:
      case JNAE:
      case JNG:
      case JNGE:
      case JNO:
      case JNS:
      case JNZ:
      case JO:
      case JPE:
      case JPO:
      case JS:
      case JZ:
      case LOOP:
      case LOOPNZ:
      case LOOPZ:
         myfputs(hnumber(atox(&instr[index]),globalstr,regs.r.cs,regs.r.cs,loc));
         if ((jmp_comments) && (instruction.name >= JA)
                                                && (instruction.name <= JZ))
            strcpy(jmpstr,equiv_jmp_names[instruction.name-JA]);
         break;
      case CALL:
      case JMP:
         if ((instruction.source == SOFF16) || (instruction.source == SOFF8))
            {
            if (byte_one == 0xeb)
               myfputs("SHORT ");
            myfputs(hnumber(atox(&instr[index]),globalstr,regs.r.cs,regs.r.cs,loc));
            }
         else
            myfputs(&instr[index]);
         break;
      case CALLF:
      case JMPF:
         if (instruction.source == SADDR32)
            {
            value = (unsigned int)((source.value >> 16) & 0xffff);
            myfputs(numberh_word(value,globalstr));
            myfputs(":");
            value = (unsigned int)(source.value & 0xffff);
            myfputs(numberh_word(value,globalstr));
            }
         else
            {
            myfputs("D");
            myfputs(&instr[index]);
            }
         break;
      case SGDT:
      case SIDT:
      case LGDT:
      case LIDT:
         myfputs("Q");
         myfputs(&instr[index]);
         break;
      case RET:
      case RETF:
         if (instruction.flags & ADJSTK)
            myfputs(numberh_word((unsigned int)source.value,globalstr));
         break;
      case INT:
         if (instruction.flags & INT3)
            source.value = (UNSIGNED_LONG)3;
         myfputs(numberh_byte((unsigned int)source.value,globalstr));
         break;
      case IN:
         index += copyto(globalstr,&instr[index],',');
         myfputs(globalstr);
         if (instruction.source == SDX)
            myfputs(&instr[index]);
         else
            myfputs(numberh_byte((unsigned int)source.value,globalstr));
         break;
      case OUT:
         index += copyto(globalstr,&instr[index],',');
         if (instruction.dest == DDX)
            myfputs(globalstr);
         else
            {
            myfputs(numberh_byte((unsigned int)dest.value,globalstr));
            myfputs(",");
            }
         myfputs(&instr[index]);
         break;
      case LDS:
      case LES:
         if (show_ptrs)
            {
            index += copyto(globalstr,&instr[index],',');
            myfputs(globalstr);
            myfputs("D");
            }
         myfputs(&instr[index]);
         break;
      default:
         if (instruction.source == SDATA16)
            {
            if (instruction.name != PUSH)  /* if not push immediate */
               {
               index += copyto(globalstr,&instr[index],',');
               myfputs(globalstr);
               }

            tempmap = (regs.r.ds == regs.r.cs) ? codemap : datamap;
/*                 && (!(PEEK(datamap,(unsigned int)source.value) & (~(PROG_DATA|PROG_OFFSET|PROG_WORD))))) */

            if ( ( ((file_type == TYPE_EXE) && (highseg != lowseg))
                    || (((unsigned int)source.value >= progstart)
                          && ((unsigned int)source.value < progend)))

                  && (use_offset)
                  && (PEEK(tempmap,(unsigned int)source.value) & (PROG_DATA|PROG_WORD|PROG_OFFSET))
                  && ((!(PEEK(tempmap,(unsigned int)source.value) & (PROG_CODE|PROG_LABEL)))
                           || ((label_code_offset)
                                && ((PEEK(tempmap,(unsigned int)source.value) & (PROG_CODE|PROG_LABEL)) == (PROG_CODE|PROG_LABEL))))
                  && (instruction.name != AND) && (instruction.name != OR)
                  && (instruction.name != XOR) && (instruction.name != TEST)
                   )

               {
               myfputs("OFFSET ");
               myfputs(hnumber((unsigned int)source.value,globalstr,regs.r.ds,regs.r.cs,loc));
               }
            else
               {
               myfputs(numberh_word((unsigned int)source.value,globalstr));
               }
            if (isaprint((unsigned char)(source.value & 0xff)))
               {
               if (isaprint((unsigned char)((source.value >> 8) & 0xff)))
                  {
                  asciistr[0] =
                  asciistr[3] = '\'';
                  asciistr[1] = (unsigned char)(source.value & 0xff);
                  asciistr[2] = (unsigned char)((source.value >> 8) & 0xff);
                  asciistr[4] = '\0';
                  }
               else
                  if ((source.value & 0xff00L) == 0L)
                     {
                     asciistr[0] =
                     asciistr[2] = '\'';
                     asciistr[1] = (unsigned char)(source.value & 0xff);
                     asciistr[3] = '\0';
                     }
               }
            }
         else
            {
            myfputs(&instr[index]);
            if (instruction.source == SDATA8)
               if (isaprint((unsigned char)(source.value & 0xff)))
                  {
                  asciistr[0] =
                  asciistr[2] = '\'';
                  asciistr[1] = (unsigned char)(source.value & 0xff);
                  asciistr[3] = '\0';
                  }
            }
         break;
      }
   }

void comment_instr(hexstr,asciistr,jmpstr,temp)
   unsigned char *hexstr, *asciistr, *jmpstr;
   unsigned int temp;
   {
   if (hexcodecomments)
      {
      tabout(56);
      myfputs(hexword(&temp,globalstr));
      myfputs(" ");
      myfputs(hexstr);
      }
   if ((ascii_comments) && (asciistr[0]) && (myfputsok != -1))
      {
      if (hexcodecomments)
         myfputs(" ");
      else
         tabout(56);
      myfputs2(asciistr,fileout);
      }
   if ((jmp_comments) && (jmpstr[0]) && (myfputsok != -1))
      {
      if (hexcodecomments)
         myfputs(" ");
      else
         tabout(56);
      myfputs(jmpstr,fileout);
      }
   }

puts_dos_comments(temp,hexstr)
   unsigned int temp;
   char *hexstr;
   {
   char *p;
   int i;

   p = NULL;
   switch ((unsigned int)source.value)
      {
      case 5:
         p = "PRINT SCREEN";
         break;
      case 0x10:
         if ((regsknown.byteknown[1]) && (regvalues.r.ax.byte.ah <= 0x1d))
            p = video_fn_names[regvalues.r.ax.byte.ah];
         else
            p = "VIDEO BIOS";
         break;
      case 0x11:
         p = "GET HARDWARE INFO";
         break;
      case 0x12:
         p = "GET MEMORY SIZE";
         break;
      case 0x13:
         if ((regsknown.byteknown[1]) && (regvalues.r.ax.byte.ah <= 0x17))
            p = disk_fn_names[regvalues.r.ax.byte.ah];
         else
            p = "DISK BIOS";
         break;
      case 0x14:
         if ((regsknown.byteknown[1]) && (regvalues.r.ax.byte.ah < 4))
            p = serial_fn_names[regvalues.r.ax.byte.ah];
         else
            p = "SERIAL I/O BIOS";
         break;
      case 0x15:
         if ((regsknown.byteknown[1]) && (regvalues.r.ax.byte.ah < 4))
            p = cassette_fn_names[regvalues.r.ax.byte.ah];
         else
            if ((regsknown.byteknown[1]) && (regvalues.r.ax.byte.ah >= 0x80)
                              && (regvalues.r.ax.byte.ah <= 0x91))
               p = atbios_fn_names[regvalues.r.ax.byte.ah-0x80];
            else
               p = "CASSETTE/AT EXTENDED";
         break;
      case 0x16:
         if ((regsknown.byteknown[1]) && (regvalues.r.ax.byte.ah < 0x13))
            p = keyboard_fn_names[regvalues.r.ax.byte.ah];
         else
            p = "KEYBOARD BIOS";
         break;
      case 0x17:
         if ((regsknown.byteknown[1]) && (regvalues.r.ax.byte.ah <= 2))
            p = parallel_fn_names[regvalues.r.ax.byte.ah];
         else
            p = "PARALLEL PRINTER I/O";
         break;
      case 0x19:
         p = "WARMBOOT SYSTEM";
         break;
      case 0x1a:
         if ((regsknown.byteknown[1]) && (regvalues.r.ax.byte.ah < 8))
            p = timer_fn_names[regvalues.r.ax.byte.ah];
         else
            p = "GET/SET TIME-OF-DAY";
         break;
      case 0x1b:
         p = "KEYBOARD BREAK";
         break;
      case 0x1c:
         p = "TIMER TICK";
         break;
      case 0x20:
         p = "TERMINATE PROGRAM";
         break;
      case 0x21:
         if ((regsknown.byteknown[1]) && (regvalues.r.ax.byte.ah <= 0x68))
            p = dos_fn_names[regvalues.r.ax.byte.ah];
         else
            p = "DOS CALL";
         break;
      case 0x25:
         p = "READ DISK SECTORS";
         break;
      case 0x26:
         p = "WRITE DISK SECTORS";
         break;
      case 0x27:
         p = "TERM. & STAY RESIDENT";
         break;
      case 0x28:
         p = "DOS TIMESLICE";
         break;
      case 0x29:
         p = "OUTPUT TO SCREEN";
         break;
      case 0x2e:
         p = "EXECUTE DOS COMMAND";
         break;
      case 0x2F:
         p = "MULTIPLEX INTERRUPT";
         break;
      case 0x33:
         if ((regsknown.wordknown[0] == W_TRUE)
                   && (regvalues.r.ax.word < 0x001f))
            p = mouse_fn_names[regvalues.r.ax.word];
         else
            p = "MOUSE DRIVER";
         break;
      case 0x40:
         if ((regsknown.byteknown[1]) && (regvalues.r.ax.byte.ah <= 5))
            p = disk_fn_names[regvalues.r.ax.byte.ah];
         else
            p = "REVECTORED FLOPPY BIOS";
         break;
      case 0x67:
         if ((regsknown.byteknown[1]) && (regvalues.r.ax.byte.ah >= 0x40)
                             && (regvalues.r.ax.byte.ah <= 0x5d))
            p = emm_fn_names[regvalues.r.ax.byte.ah-0x40];
         else
            p = "EMM FUNCTIONS";
         break;
      }
   if (p != NULL)
      {
      tabout(48);
      myfputs(p);
      }
   else
      if (hexcodecomments)
         {
         tabout(56);
         myfputs(hexword(&temp,globalstr));
         myfputs(" ");
         myfputs(hexstr);
         }
   }

tabout(maxlen)
   int maxlen;
   {
   if (myfputsok != -1)
      {
      if (fileout_len >= maxlen)
         {
         myfputc(' ',fileout);
         myfputc(';',fileout);
         }
      else
         {
         if (use_tabs)
            {
            fileout_len -= fileout_len % 8;
            do
               {
               fileout_len += 8;
               myfputc('\t',fileout);
               } while (fileout_len < maxlen);
            }
         else
            for ( ; fileout_len < maxlen; fileout_len++)
               myfputc(' ',fileout);
         myfputc(';',fileout);
         }
      }
   }

int check_str()  /* uses global strings */
   {
   int i,j;
   char *p;
   char ch,ch2;  /* ch2 to keep stack aligned */

   j = strlen(search_str);
   ch = search_str[0];
   if (find_str_len > j)
      for (i = 0, p = find_str; i < find_str_len-j; i++, p++)
         if (*p == ch)
            if (strncmp(p,search_str,j) == 0)
               return(TRUE);
   return(FALSE);
   }

int myfputc(ch,fileout)
   char ch;
   int fileout;
   {
   if (string_search)
      {
      find_str[find_str_len++] = ch;
      if (ch == '\n')
         {
         find_str[find_str_len] = '\0';  /* string wasn't terminated before */
         strlwr(find_str);
         if (check_str())
            {
            string_found = TRUE;
            puts(hexword(&search_ip,globalstr));
            puts(":");
            puts(nl_str);
            puts(find_str);
            }
         find_str_len = 0;
         }
      return(TRUE);
      }
   else
      return(fputc(ch,fileout));
   }

void myfputs2(str)
   char *str;
   {
   int temp;
   char ch,ch2;  /* 2 to keep stack even */

   if (myfputsok != -1)
      {
      while (ch = *(str++))
         {
         fileout_len++;
         if (ch == '\t')
            {
            temp = fileout_len;
            fileout_len += 7 - ((fileout_len-1) % 8);
                                               /* -1 because of inc above */
            if (!use_tabs)
               {
               for (; temp < fileout_len; temp++)
                  myfputc(' ',fileout);
               ch = ' ';
               }
            }
         if (ch == '\n')
            {
            fileout_len = 0;
            if ((pagepause) && (++linecount > 23))
               {
               if (pressany(TRUE) == '\033')
                  {
                  myfputsok = -1;  /* if Esc, quit! */
                  break;
                  }
               linecount = 0;
               }
            myfputc('\r',fileout);
            }
         myfputsok = myfputc(ch,fileout);
         }
      }
   }

void myfputs(str)
   char *str;
   {
   int temp;
   char ch,ch2;  /* 2 to keep stack even */

   if (myfputsok != -1)
      {
      if (!upperflag)
         {
         strcpy(lowerstr,str);  /* do not lowercase constant strings! */
         str = strlwr(lowerstr);
         }
      while (ch = *(str++))
         {
         fileout_len++;
         if (ch == '\t')
            {
            temp = fileout_len;
            fileout_len += 7 - ((fileout_len-1) % 8);
                                               /* -1 because of inc above */
            if (!use_tabs)
               {
               for (; temp < fileout_len; temp++)
                  myfputc(' ',fileout);
               ch = ' ';
               }
            }
         if (ch == '\n')
            {
            fileout_len = 0;
            if ((pagepause) && (++linecount > 23))
               {
               if (pressany(TRUE) == '\033')
                  {
                  myfputsok = -1;  /* if Esc, quit! */
                  break;
                  }
               linecount = 0;
               }
            myfputc('\r',fileout);
            }
         myfputsok = myfputc(ch,fileout);
         }
      }
   }

int copyto(dest,src,stop)
   char *dest, *src;
   char stop;
   {
   int num;

   num = 0;
   do
      {
      dest[num] = src[num];
      } while ((src[num]) && (src[num++] != stop));
   dest[num] = '\0';
   return(src[num-1] == stop ? num : -1);
   }

int numval(str,val)  /* change hex value in str to a number */
                     /* return FALSE if invalid hex number, else TRUE */
   unsigned char *str;
   unsigned int *val;  /* Beware of passing an INT pointer to this routine
                           if the machine does not use low-byte first! */
   {
   int i,valid;
   unsigned int ch;

   i = 0;
   *val = 0;
   valid = FALSE;
   ch = toupper(str[i++]);
   while (ch)
      {
      valid = TRUE;
      if (isdigit(ch))
         *val = (*val << 4) + ch - '0';  /* a number */
      else
         if ((ch >= 'A') && (ch <= 'F'))  /* a letter */
            *val = (*val << 4) + ch - 'A' + 10;
         else
            return(FALSE);
      ch = toupper(str[i++]);
      }
   return(valid);
   }

int load_file()  /* determine if an .EXE load has to be done, else load
                    in .COM (data) format */
   {
   if (file_type == TYPE_EXE)
      return(load_exe());
   return(load_com());
   }

int load_com()  /* I will try to make this transportable, but it will
                    probably wind up DeSmet specific */
   {
   int filein;
   unsigned int numin,seg,ofs,seg2,ofs2,count;

   filein = open(filename,MODE_READ);  /* should be able to open it again */
   ofs = load_offset & 0x000f;
   seg = (load_offset >> 4) & 0x0fff;
   checkaddress(&seg,&ofs);  /* translate address */
   do
      {
      numin = read(filein,diskbuf,DISKBUF_SIZE);
      if (numin)
         {
         MEMCPY(numin,datasegment,diskbuf,seg,ofs);
         seg += (numin >> 4) & 0x0fff;  /* add offset for next read */
         ofs += numin & 0x000f;
         }
      } while (numin);
   close(filein);
   return(TRUE);
   }

int load_exe()  /* load and relocate as necessary an .EXE file */
   {
   struct imageheader
      {
      unsigned int exesignature;
      unsigned int lenmod512;
      unsigned int sizeinpages;
      unsigned int numofrelocitems;
      unsigned int headersize;
      unsigned int minmem;
      unsigned int maxmem;
      unsigned int ssdisp;
      unsigned int spvalue;
      unsigned int checksum;
      unsigned int ipvalue;
      unsigned int csdisp;
      unsigned int relocdisp;
      unsigned int ovlynumber;
      } header;
   unsigned int relseg,relofs,tableindex,numreloc,seg,ofs,numin,temp;
   UNSIGNED_LONG minmem,filepos,filelen,tablepos;
   int filein;

   filein = open(filename,MODE_READ);  /* should be able to open it again */
   numin = read(filein,&header,sizeof(struct imageheader));
                                                  /* read initial header */
   filepos = ((long)header.headersize) << 4;
   ofs = load_offset & 0x000f;
   seg = (load_offset >> 4) & 0x0fff;
   regs.r.cs = header.csdisp;  /* set registers */
   regs.r.ss = reloc_ds;
   regs.r.sp = header.spvalue;
   regs.r.ip = header.ipvalue;
   checkaddress(&seg,&ofs);  /* translate address */
   memok = TRUE;  /* no trapping here */
   lseek(filein,filepos,0);  /* seek past header */
   do
      {
      numin = read(filein,diskbuf,DISKBUF_SIZE);  /* read this chunk */
      if (numin)
         {
         MEMCPY(numin,datasegment,diskbuf,seg,ofs);
         seg += (numin >> 4) & 0x0fff;  /* paragraphs in */
         ofs += numin & 0x000f;  /* bytes in */
         }
      } while (numin);

   close(filein);
   return(TRUE);
   }

int get_reloc_params(line)
   int line;
   {
   struct imageheader
      {
      unsigned int exesignature;
      unsigned int lenmod512;
      unsigned int sizeinpages;
      unsigned int numofrelocitems;
      unsigned int headersize;
      unsigned int minmem;
      unsigned int maxmem;
      unsigned int ssdisp;
      unsigned int spvalue;
      unsigned int checksum;
      unsigned int ipvalue;
      unsigned int csdisp;
      unsigned int relocdisp;
      unsigned int ovlynumber;
      } header;
   union tableunion
      {
      unsigned char longstr[512];
      struct tables
         {
         unsigned int relocofs;
         unsigned int relocseg;
         } table[128];
      } tbl;
   unsigned int relseg,relofs,tableindex,numreloc,seg,ofs,numin,temp;
   UNSIGNED_LONG minmem,filepos,filelen,tablepos, headerbytes;
   int filein;

   multi_segment = FALSE;
   reloc_ds = 0;
   filein = open(filename,MODE_READ);  /* should be able to open it again */
   filelen = lseek(filein,0L,2);
   lseek(filein,0L,0);  /* go back to start */
   if (file_type == TYPE_EXE)
      {
      numin = read(filein,&header,sizeof(struct imageheader));
                                                  /* read initial header */
      headerbytes = ((long)header.headersize) << 4;
      filelen -= headerbytes;
      if ((filelen / 16) > maxpara)  /* if too big */
         {
         putsat(line,0,"Error: not enough memory.");
         pressany(FALSE);
         close(filein);
         return(FALSE);
         }
      seg = 0;  /* variables for file load */
      ofs = load_offset;
      regs.r.cs = header.csdisp;  /* set registers */
      recom_ds =
      regs.r.ss =
      regs.r.ds =
      regs.r.es = header.ssdisp;
      initial_sp =
      regs.r.sp = header.spvalue;
      startloc =
      regs.r.ip = header.ipvalue;
      numin = 0;
      tablepos = header.relocdisp;  /* this changes it to a long */
      lseek(filein,tablepos,0);
      while (numin < header.numofrelocitems)  /* check program segmentation */
         {
         if ((numin % 128) == 0)  /* if an empty relocation table */
            {
            tableindex = 0;
            read(filein,tbl.longstr,512);
                           /* read more of the table */
            }
         filepos = ((long)tbl.table[tableindex].relocseg) << 4;
         filepos += (long)tbl.table[tableindex++].relocofs;
         lseek(filein,filepos+headerbytes,0);
         read(filein,&temp,2);  /* read reloc info */
         if (temp)
            {
            if (!reloc_ds)
               reloc_ds = temp;
            else
               if (reloc_ds != temp)
                  {
                  multi_segment = TRUE;
                  break;
                  }
            }
         numin++;
         }

      if (reloc_ds)
         {
         recom_ds =
         regs.r.ss = reloc_ds;
         }
      else
         reloc_ds = recom_ds;
      close(filein);
      if (multi_segment)
         {
         putsat(line,0,"WARNING:  This .EXE file is multi-segmented, and cannot be disassembled\n");
         puts("by CODEGEN.  Please try another file.");
         pressany(FALSE);
         return(FALSE);
         }
/*
curpos(20,0);
printf("ss:sp = %04x:%04x   cs:ip = %04x:%04x \n",regs.r.ss,regs.r.sp,
regs.r.cs,regs.r.ip);
*/
      if (regs.r.ss > regs.r.cs)
         {
         lowseg = regs.r.cs;
         highseg = regs.r.ss;
         }
      else
         {
         lowseg = regs.r.ss;
         highseg = regs.r.cs;
         }
      if ((highseg - lowseg) > 0x0fff)
         {
         putsat(line,0,"Error: Segment > 64K!");
         pressany(FALSE);
         return(FALSE);
         }
/*
printf("lowseg=%04x   highseg=%04x   load_offset=%04x  filelen=%04lx\n",
lowseg,highseg,load_offset,filelen);
*/
      progstart = load_offset;
      physical_eof =
      progend = filelen - ((long)highseg << 4);
/*
printf("filelen=%04lx  highseg=%04x   progend=%04x   ((long)highseg<<4)=%04lx\n",
filelen,highseg,progend,((long)highseg<<4));
*/
/*
      if (regs.r.cs == lowseg)
         progend = (highseg-lowseg) << 4;
*/
      if (highseg - lowseg)
         datamap = mapspace + 0x1000;
      else
         {
         progend = (filelen > 0xfffeL) ? 0xfffe : filelen;
         progend -= ((long)highseg << 4);
         datamap = mapspace;  /* reset, can change workfile! */
/*         file_type = TYPE_COM; */  /* this doesn't look like an .EXE file */
         memory_model = 2;
         }
      if (regs.r.ss != highseg)  /* if the data comes first */
         {
         temp = datamap;
         datamap = codemap;
         codemap = temp;
         }
      }
   else   /* COM or DEV type */
      {
      close(filein);
      if ((filelen > 0xffffL) || (load_offset > load_offset+(unsigned int)filelen))
         {
         putsat(line,0,"Error: Program too big!");
         pressany(FALSE);
         return(FALSE);
         }
      progstart = load_offset;
      physical_eof =
      progend = load_offset + (unsigned int)filelen;
      }
/*
printf("startloc=%04x, progstart=%04x, progend=%04x  ",startloc,progstart,progend);
ci();
*/
   if ((startloc < progstart) || (startloc > progend))
      {
      putsat(line,0,bad_start_ip_msg);
/*
printf("startloc=%04x   progstart=%04x   progend=%04x\n",startloc,progstart,progend);
*/
      startloc = load_offset;
      pressany(FALSE);
      }
   return(TRUE);
   }


void checkaddress(seg,ofs)  /* check memory, ignore virtual screen */
   unsigned int *seg, *ofs;
   {
   unsigned int t1,aseg,count;

   aseg = (t1 = *seg + (*ofs >> 4)) >> 12;
   *seg += memmap[aseg];
   }

void setmaps()
   {
   unsigned int count;

   memtop = peekw(0,0x413);  /* get system memory */
   usedmem = mapspace + 0x4000;  /* top of memory used */
   availmem = ((memtop << 6) - usedmem >= 0x1000) ? TRUE : FALSE;
                             /* how much is left? */
   usedmem -= pspaddr;   /* paragraphs to keep during shell */
   memtop = ((memtop >> 6) & 0x000f) + ((memtop & 0x3f) ? 1:0);
                                                  /* get in low nibble */
   memused = (segoffset >> 12) & 0x000f;
                                      /* get number of 64K blocks used */
   maxpara = (memtop - memused) * 0x1000;  /* number of paragraphs in
                                              virtual memory */
   for (count = 0; count < 16; count++)
      {
      if (count >= (memtop - memused))
         {
         if (count > 9)
            memmap[count] = 0;  /* direct memory mapping */
         else
            memmap[count] = deadmem - (0x1000 * count);
                                           /* map to deadmem:xxxx */
         }
      else
         memmap[count] = segoffset;  /* map to virtual memory */
      }
   }

dos_shell()
   {
#asm
   mov es,word pspaddr_         ;get PSP segment
   mov bx,word usedmem_         ;get desired mem size
   mov ah,4ah                   ;modify block size to bx
   int 21h                      ;call DOS
   mov ax,ds                    ;get current data segment
   mov word dos4b_block_[4],ax  ;save cmdline seg
   mov word dos4b_block_[8],ax  ;save FCB1 seg
   mov word dos4b_block_[12],ax ;save FCB2 seg
   mov ax,word es:[2ch]         ;get environment seg
   mov word dos4b_block_,ax     ;save ptr to env seg
   or ax,ax                     ;have an env?
   jz std_dos_shell             ;no, skip this
   mov es,ax                    ;env seg to es
   cld                          ;up
   xor di,di                    ;start at 0 in env
ds_srch_env_loop:
   cmp byte es:[di],0           ;end of string?
   je ds_check_end_env          ;yes, see if end of env
   push di                      ;save start of string
   mov si,offset comspec_name_  ;name to match
   mov cx,8                     ;len of "COMSPEC="
   db 0f3h                      ;REPE
   cmpsb                        ;compare strings
   pop bx                       ;get start of string back
   je ds_have_comspec           ;and go set comspec
   mov di,bx                    ;get ptr to start of string
   xor al,al                    ;get a 0
   mov cx,7fffh                 ;a big number
   db 0f3h                      ;REP
   scasb                        ;find a 0
   dec di                       ;back to the 0
ds_check_end_env:
   inc di                       ;next byte
   cmp byte es:[di],0           ;end of environment?
   jne ds_srch_env_loop         ;check next string
std_dos_shell:
   push ds                      ;current data segment
   pop es                       ;to es for our use
   mov di,offset command_name_  ;default COMMAND.COM name
ds_have_comspec:
   mov word cs:ds_ss_save,ss    ;save stack/data segment
   mov word cs:ds_sp_save,sp    ;save stack pointer
   push es                      ;name data segment
   push ds                      ;and current data segment
   pop es                       ;to seg of parameter block
   pop ds                       ;and seg of name
   mov bx,offset dos4b_block_   ;es:bx = ptr to parameter block
   mov dx,di                    ;ds:dx = ptr to name to execute
   mov ax,4b00h                 ;load and execute program
   int 21h                      ;call DOS
   cli                          ;do not disturb!
   mov ss,word cs:ds_ss_save    ;get saved stack segment
   mov sp,word cs:ds_sp_save    ;and saved stack pointer
   sti                          ;interrupts are OK again
   push ss                      ;copy stack/data segment
   pop ds                       ;to ds and done
   pop bp                       ;restore bp
   ret                          ;and done
ds_ss_save: dw 0                ;SS save area
ds_sp_save: dw 0                ;SP save area
#endasm
   }
