/*
   This file is copyright (c) 1986 by William Couture.  All rights reserved.
*/


/*
   General header file associated with the files in the VIM program.

   Note that STDIO is not used.  Place all such common definitions
   in this file instead.

   Modifed for CODEGEN
*/

/* masks to set appropriate flags in the flags register by ORing */
#define CARRY 0x0001
#define PARITY 0x0004
#define AUXCARRY 0x0010
#define ZERO 0x0040
#define SIGN 0x0080
#define TRAP 0x0100
#define INTERRUPT 0x0200
#define DIRECTION 0x0400
#define OVERFLOW 0x0800

/* masks to clear appropriate flags in the flags register by ANDing */
#define CARRYOFF 0xfffe
#define PARITYOFF 0xfffb
#define AUXCARRYOFF 0xffef
#define ZEROOFF 0xffbf
#define SIGNOFF 0xff7f
#define TRAPOFF 0xfeff
#define INTERRUPTOFF 0xfdff
#define DIRECTIONOFF 0xfbff
#define OVERFLOWOFF 0xf7ff

#define SIGNEDB 0x80  /* mask for sign bit in 8 bit quantity */
#define SIGNEDW 0x8000  /* mask for sign bit in 16 bit quantity */

#define MEMORY 0x0000  /* source/destination is memory */
#define REGISTER 0x0001  /* source/destination is a register */
#define SEGREG 0x0002  /* source/destination is a segment register */
#define REG 0x0003  /* source/destination is a register or segment register */
#define MEM_REG_UNKNOWN 0xf0  /* is unknown */
   /* Note:  I seem to have used REG a lot where I probably should have used
      REGISTER (in fns.c).  Everything seems to work, though. */

#define ESSEG 0x0001  /* prefix is ES segment override */
#define CSSEG 0x0002  /* prefix is CS segment override */
#define SSSEG 0x0004  /* prefix is SS segment override */
#define DSSEG 0x0008  /* prefix is DS segment override */
#define SEG   0x000f  /* prefix is one of the segment overrides */
#define REPZERO  0x0010  /* prefix is REPZ/REPE */
#define REPNOTZERO 0x0020  /* prefix is REPNZ/REPNE */
#define REP   0x0030  /* prefix is one of the REP instructions */
#define FORGETZF 0x0100  /* Set only by string primitive instruction in
                            fns.c, says what this primitive does with a
                            REPxx instruction */
#define ISZFON 0x0200  /* as above for FORGETZF */
#define ISZFOFF 0x0400  /* as above for FORGETZF */
#define LOCK_PREFIX 0x8000  /* for CODEGEN, say we have a LOCK prefix */

#define HALTERR 1  /* just executed a HALT instruction */
#define NOHALT 0

/*
#define P3B4 0  define video port addresses in the mdaports array
#define P3B5 1
#define P3B8 2
#define P3B9 3
#define P3BA 4
#define P3BB 5
#define P3BF 6

#define P3D4 0  define video port addresses in the cgaports array
#define P3D5 1
#define P3D8 2
#define P3D9 3
#define P3DA 4
#define P3DB 5

#define P3C2 0  define the video port addresses in the egaports array
#define P3xA 1
#define P3C4 2
#define P3CC 3
#define P3CA 4
#define P3CE 5
#define P3C0 6
*/

#define TRUE 1  /* standard true */
#define FALSE 0  /* standard false */

#define YES 1  /* convenient definitions */
#define NO 0

#define HWINTBASE 8  /* lowest numbered hardware interrupt */
#define ATINTBASE 0x70  /* the lowest numbered new AT interrupt */
#define ATINTOFFSET 104 /* 0x68, AT interrupts - HWINTBASE (used to
                           position the flag in hwintstack) */

#define ESC_KEY '\033'  /* the ESC key */

#define SYM_TABLE_SIZE 512

#define REGS_UNKNOWN _setmem(&regsknown.byteknown[0],28,0)

#define NULL 0
#define W_TRUE 0x0101
/* true in both high and low bytes */

/*
   Structure definitions
*/

struct optable  /*  Define the optable structure */
   {
   int name;  /* "name" is also location in arrays */
   unsigned int flags;  /* conditions associated with this instruction */
   int dest;  /* to where (mem/reg/etc) */
   int source;  /* from where (mem/reg/etc) */
   };


struct tofrom  /* stuff associated with instruction destination and source */
   {
   UNSIGNED_LONG value;  /* value of source/destination */
   UNSIGNED_LONG oldvalue;  /* value before it is changed by instruction,
                               used to set flags */
       /* Note:  The file COMPILER.H must be included BEFORE VIM.H, so that
          UNSIGNED_LONG has been TYPEDEFed */
   unsigned int where;  /* is the source/destination register or memory? */
   unsigned int reg;  /* register for source/destination */
   unsigned int seg;  /* segment for souce/destination */
   unsigned int ofs;  /* offset for source/destination */
   };

struct addrtrap            /* structure of the memory address trap info */
   {
   UNSIGNED_LONG lowaddr;  /* the lower bound */
   UNSIGNED_LONG highaddr;  /* the upper bound */
   };

struct intrptr        /* interrupt pointer, for internal use to watch
                         the hardware interrupts */
   {
   unsigned int ptrofs;
   unsigned int ptrseg;
   };

struct symbol_table
   {
   unsigned int ofs;   /* offset of the label */
   unsigned int seg;   /* segment of the label */
   unsigned char sym[32];  /* and a 31-byte (plus terminator) symbol */
   };

struct cgmap_header
   {
   char signature[4];
   unsigned int map_ofs;  /* ofs of start of map in file */
   unsigned int load_ofs;
   unsigned int load_seg;
   unsigned int filetype_flag;
   unsigned int exe_model;
   unsigned int exe_data_seg;
   unsigned int flagword;
   unsigned int start_offset;  /* always significant */
   };

#define CGM_FLAG_RESERVED1 0x0001
            /* CGM_FLAG_DOMAP no longer used, but needs to be reserved
               for compatibility with older CGM files */
#define CGM_FLAG_INTADDR   0x0002
#define CGM_FLAG_PROCESSOR 0x0004
#define CGM_FLAG_MOREFLAGS 0x8000
            /* flag additional flagwords */

#define TYPE_COM 0
#define TYPE_EXE 1
#define TYPE_DEV 2

struct call_addr
   {
   unsigned int offset;
   unsigned char endbyte_flag;
   unsigned char endbyte;
   };

struct int_offset
   {
   unsigned char intnum;
   unsigned int intofs;
   };

#define NUM_NORET_CALL_ADDRS 10
#define NUM_ROM_INTS 72
        /* allow for extra INT from BIOS entry points */


#define SETFLAG(x,f)   _poke(_peek(x,mapspace) | (f),x,mapspace)
#define SETCFLAG(x,f)  _poke(_peek(x,codemap) | (f),x,codemap)
#define SETDFLAG(x,f)  _poke(_peek(x,datamap) | (f),x,datamap)

#define PROG_UNKNOWN  0x00   /* unknown byte, possibly data */
#define PROG_CODE     0x01   /* unlabeled Code */
#define PROG_DATA     0x02   /* Data (labeled, of course!) */
#define PROG_LABEL    0x04   /* code is labeled */
#define PROG_WORD     0x08   /* byte or word reference for data */
#define PROG_OFFSET   0x10   /* could be a label */
#define PROG_ADDRESS  0x40   /* location contains a label */
#define PROG_KNOWN    0x80   /* has been disassembled */

#define PROG_CRUNCHED 0x7f  /* crunched for map read/write
                               SOME of these bits have to be exclusive! */

#define DISKBUF_SIZE 4096
/* #define NUM_INDEX_ENTRIES 340 */
#define INDEX_ENTRIES_BYTES 4080

struct xref_index_struc
   {
   unsigned int seg;  /* segment being referenced */
   unsigned int ofs;  /* offset being referenced */
   long first;        /* location of first entry */
   long last;         /* location of last entry (to set NEXT link) */
   };

struct xref_entry_struc
   {
   unsigned int seg;  /* segment of reference */
   unsigned int ofs;  /* offset of reference */
   long next;         /* ptr to next reference */
   };
