/*
   This file is copyright (c) 1986 by William Couture.  All rights reserved.
*/


/*
   COMPILER.H - Defines (most) of the compiler-specific functions used
                in the VIM program.  Most of the work of porting the
                program will be to simply change this file.

                NOTE:  Due to such things as interrupt handling routines,
                etc., INLINE ASSEMBLER had to be used!  This is NOT
                affected by changing this file, nor is the IMPLICIT and
                EXPLICIT use of the SMALL MEMORY MODEL.  See VIM.C for
                an explanation of the use of the small memory model.

                Also, BEWARE of the byte-order problem when changing CPU's!
                In particular, the register unions and the hexbyte(),
                hexword(), and hexlong() functions all assume low-byte
                first, and there may be others.


      *** THE COMPILER SUPPORTED BY THIS VERSION OF COMPILER.H IS ***

                          *** DeSmet V2.51 ***
*/

#define PEEK(s,o) _peek(o,s)
    /* I like things in seg,offset format, because it seems more natural
       (after having used assembly language, debug, etc) to specify things
       as seg:offset. */
#define POKE(c,s,o) _poke(c,o,s)  /* character,segment,offset */

#define OUTBPORT(v,a) _outb(v,a)
               /* send value V to byte output port at address A */
#define OUTWPORT(v,a) _outw(v,a)
               /* send value V to word output port at address A */
#define INBPORT(a) _inb(a)
           /* get a byte value from the input port at address A */
#define INWPORT(a) _inw(a)
           /* get a word value from the input port at address A */

#define MEMCPY(n,ss,so,ts,to) _lmove(n,so,ss,to,ts)
                   /* Note that n is an integer, and thus can only be
                      0-65535!  The other variables are ss: source seg,
                      so: source offset, ts: to seg, to: to offset */
#define DATASEG _showds()  /* The function to get the data segment is called
                              _showds(). */
#define CODESEG _showcs()  /* The function to get the code segment is called
                              _showcs(). */
    /* Note that these functions imply a SINGLE data segment and a SINGLE code
       segment, i.e., a small memory model. */
#define PUTS(s) myputs(s)  /* These programs use the DeSmet style for puts(),
                  which does not add a \n on the end of a string.  Since this
                  makes it handy for formatting, you will have to find/write
                  the equivalent for another compiler.  Also, the special
                  version of puts() used here allows I/O redirection inside
                  VIM that is invisible to the virtual DOS. */
#define GETS(s) mygets(s)  /* a special version of gets() to allow I/O
                              re-direction invisible to the virtual DOS. */

#define ERR -1  /* Err return for file open */
#define MODE_READ 0  /* for call to open() */
#define READ_WRITE 2  /* for call to open() */

typedef long UNSIGNED_LONG;  /* I want unsigned longs (i.e. 32-bit booleans),
                                but I will have to settle for this */
