/*
   This file is copyright (c) 1986, 1987, 1990 by William Couture.  All rights reserved.
*/


/*
   Defines for the instruction names.

   NOTE:  The order of defines (and their values) is also keyed to the
   opnames array (holds the names of the instructions) and the opfns
   array (holds the pointers to functions).
   IF YOU CHANGE ANY OF THESE, YOU NEED TO CHANGE THEM ALL!
*/


#define UNDEF -3   /* not a valid instruction */
#define PMBYTE -2  /* need 80286 byte 2 table */
#define UNKNOWN -1  /* need the second byte to finish decoding */
#define DBYTE 0  /* a fence value to separate valid from invalid */
#define AAA 1  /* the 8088 instructions in alaphabetical order */
#define AAD 2
#define AAM 3
#define AAS 4
#define ADC 5
#define ADD 6
#define AND 7
#define BOUND 8
#define CALL 9
#define CALLF 10  /* for CALL FAR ... */
#define CBW 11
#define CLC 12
#define CLD 13
#define CLI 14
#define CMC 15
#define CMP 16
#define CMPSB 17
#define CMPSW 18
#define CWD 19
#define DAA 20
#define DAS 21
#define DEC 22
#define DIV 23
#define ENTER 24
#define ESC 25
#define HLT 26
#define IDIV 27
#define IMUL 28
#define IN 29
#define INC 30
#define INSB 31
#define INSW 32
#define INT 33
#define INTO 34
#define IRET 35
#define JA 36
#define JAE 37
#define JCXZ 38
#define JG 39
#define JGE 40
#define JMP 41
#define JMPF 42  /* for JMP FAR ... */
#define JNA 43
#define JNAE 44
#define JNG 45
#define JNGE 46
#define JNO 47
#define JNS 48
#define JNZ 49
#define JO 50
#define JPE 51
#define JPO 52
#define JS 53
#define JZ 54
#define LAHF 55
#define LDS 56
#define LEA 57
#define LEAVE 58
#define LES 59
#define LOCK 60
#define LODSB 61
#define LODSW 62
#define LOOP 63
#define LOOPNZ 64
#define LOOPZ 65
#define MOV 66
#define MOVSB 67
#define MOVSW 68
#define MUL 69
#define NEG 70
#define NOP 71
#define NOT 72
#define OR 73
#define OUT 74
#define OUTSB 75
#define OUTSW 76
#define POP 77
#define POPA 78
#define POPF 79
#define PUSH 80
#define PUSHA 81
#define PUSHF 82
#define RCL 83
#define RCR 84
#define REPZ 85
#define REPNZ 86
#define RET 87
#define RETF 88  /* for RET FAR */
#define ROL 89
#define ROR 90
#define SAHF 91
#define SAR 92
#define SBB 93
#define SCASB 94
#define SCASW 95
#define SEGCS 96
#define SEGDS 97
#define SEGES 98
#define SEGSS 99
#define SHL 100
#define SHR 101
#define STC 102
#define STD 103
#define STI 104
#define STOSB 105
#define STOSW 106
#define SUB 107
#define TEST 108
#define WAIT 109
#define XCHG 110
#define XLAT 111
#define XOR 112
/*
      The 80286 specific instructions are tacked on to the end
      in no specific order.
*/
#define ARPL 113
#define SLDT 114
#define STR 115
#define LLDT 116
#define LTR 117
#define VERR 118
#define VERW 119
#define SGDT 120
#define SIDT 121
#define LGDT 122
#define LIDT 123
#define SMSW 124
#define LMSW 125
#define LAR 126
#define LSL 127
#define CLTS 128



/* Flagword defines.  They say that this instruction has the associated
   special properties associated with it. */

#define UNK 0x0000  /* need the second byte to finish decoding */
#define NONE 0x0000  /* no properties */
#define BYTE 0x0001  /* byte instruction */
#define WORD 0x0002  /* word instruction */
#define NOCOUNT 0x0004  /* a shr/ror/rcl/etc. instruction with count of 1 */
#define COUNT 0x0008  /* a shr/ror/rcl/etc. instruction with a count of CL */
#define DATA8 0x0010  /* Immediate 8 bit data after instruction */
#define DATA16 0x0020  /* Immediate 16 bit data after instruction */
#define XDATA8 0x0040  /* Immediate sign-extended 8 bit data after instr. */
#define OFF8 0x0080  /* Immediate 8-bit signed offset after instruction */
#define OFF16 0x0100  /* Immediate 16-bit signed offset after instruction */
#define ADDR16 0x0200  /* Immediate 16 bit address after instruction */
#define ADDR32 0x0400  /* Immediate 32 bit address after instruction */
#define INT3 0x0800  /* Type 3 Interrupt */
#define AADAAM 0x1000  /* The AAD or AAM instruction - special decoding */
#define ADJSTK 0x2000  /* Adjust stack after return by immediate 16 bit
                          quantity */
#define PREFIX 0x4000  /* Instruction is a prefix instruction */
#define ACCUM 0x8000  /* Associated register is AL or AX - accumulator
                         instruction */
#define ACCUM8 0x8001  /* AL accumulator instruction */
#define ACCUM16 0x8002  /* AX accumulator instruction */

/* Destination defines */

/*
   Note:  At one time, the Destination and Source defines had to be identical.
   I'm not sure this is true anymore, but if you are going to change them,
   check it out.

   For more information on where these values came from and what they
   represent, check out the 8086 & 8088 instant reference card mentioned
   in COMMENTS.H.
*/

#define DZ 32000  /* No destination (i.e. jmp, loop, etc., instruction) */
    /* just a big number to be above everything else */
#define DA 1  /* EA byte needed to decode furthur, look in Table A */
#define DB 2  /* Table B */
#define DC 3  /* Table C */
#define DD 4  /* Table D */
#define DE 5  /* Table E */
#define DF 6  /* Table F */
#define DG 7  /* Table G */
#define DH 8  /* Table H */
#define DI 9  /* Table I */
#define DJ 10  /* Table J */
#define DK 11  /* Table K */
#define DS 12  /* Table S (80286 instructions) */
#define DT 13  /* Table T (80286 instructions) */
#define DU 14  /* Table U (80286 PM instructions) */
#define DV 15  /* Table V (80286 PM instructions) */
#define DL 16  /* Destination is segment register */
#define DM 17  /* Destination is a memory reference */
#define DDATA8 18  /* Destination is 8 bit data (not used) */
#define DDATA16 19  /* Destination is 16 bit data */
#define DXDATA8 20  /* Destination is sign extended 8 bit data (not used) */
#define D8 21  /* Destination is 8 bit register */
#define D16 22  /* Destination is 16 bit register */
#define DAX 64  /* Destination is AX register */
   /* Note that register destinations are in opcode order */
#define DCX 65
#define DDX 66
#define DBX 67
#define DSP 68
#define DBP 69
#define DSI 70
#define DDI 71
#define DES 72
#define DCS 73
#define DSS 74
#define DDS 75
#define DFL 76
   /* end of opcode order for a moment */
#define DRAL 77  /* Accumulator AL */
#define DRAX 78  /* Accumulator AX */
   /* start of opcode order again */
#define DAL 79  /* AL */
#define DCL 80
#define DDL 81
#define DBL 82
#define DAH 83
#define DCH 84
#define DDH 85
#define DBH 86
#define DOFF8 256  /* Destination is 8 bit offset (not used) */
#define DOFF16 257  /* Destination is 16 bit offset (not used) */
#define DADDR16 258  /* Destination is 16 bit address */
#define DADDR32 259  /* Destination is 32 bit address (not used) */


/* Source defines */

#define SZ 32000  /* all comments for destination defines apply, excpet that
                     ALL source defines are used */
#define SA 1
#define SB 2
#define SC 3
#define SD 4
#define SE 5
#define SF 6
#define SG 7
#define SH 8
#define SI 9
#define SJ 10
#define SK 11
#define SS 12
#define ST 13
#define SU 14
#define SV 15
#define SL 16
#define SM 17
#define SDATA8 18
#define SDATA16 19
#define SXDATA8 20
#define S8 21
#define S16 22
#define SAX 64
#define SCX 65
#define SDX 66
#define SBX 67
#define SSP 68
#define SBP 69
#define SSI 70
#define SDI 71
#define SES 72
#define SCS 73
#define SSS 74
#define SDS 75
#define SFL 76
#define SRAL 77
#define SRAX 78
#define SAL 79
#define SCL 80
#define SDL 81
#define SBL 82
#define SAH 83
#define SCH 84
#define SDH 85
#define SBH 86
#define SOFF8 256
#define SOFF16 257
#define SADDR16 258
#define SADDR32 259
